# coding=utf-8
""" Tools for Earth Engine ee.List objects """

import ee
from . import computedobject


def difference(eelist, to_compare):
    """ Difference between two earth engine lists

    :param ee_list2: the other list
    :return: list with the values of the difference
    :rtype: ee.List
    """
    return eelist.removeAll(to_compare).add(to_compare.removeAll(eelist)) \
        .flatten()


def format(eelist):
    """ Convert a list to a string """
    def wrap(el, ini):
        ini = ee.String(ini)
        strel = ee.Algorithms.String(el)
        return ini.cat(',').cat(strel)

    liststr = ee.String(eelist.iterate(wrap, ''))
    return liststr.replace('^,', '[').cat(']')


def getFromDict(eelist, values):
    """ Get a list of Dict's values from a list object. Keys must be unique

    :param values: dict to get the values for list's keys
    :type values: ee.Dictionary
    :return: a list of values
    :rtype: ee.List
    """
    values = ee.Dictionary(values) if isinstance(values, dict) else values

    empty = ee.List([])

    def wrap(el, first):
        f = ee.List(first)
        cond = values.contains(el)
        return ee.Algorithms.If(cond, f.add(values.get(el)), f)

    values = ee.List(eelist.iterate(wrap, empty))
    return values


def intersection(eelist, intersect):
    """ Find matching values. If ee_list1 has duplicated values that are
    present on ee_list2, all values from ee_list1 will apear in the result

    :param intersect: the other Earth Engine List
    :return: list with the intersection (matching values)
    :rtype: ee.List
    """
    eelist = ee.List(eelist)
    intersect = ee.List(intersect)
    newlist = ee.List([])
    def wrap(element, first):
        first = ee.List(first)

        return ee.Algorithms.If(intersect.contains(element),
                                first.add(element), first)

    return ee.List(eelist.iterate(wrap, newlist))


def removeDuplicates(eelist):
    """ Remove duplicated values from a EE list object """
    # TODO: See ee.List.distinct()
    newlist = ee.List([])
    def wrap(element, init):
        init = ee.List(init)
        contained = init.contains(element)
        return ee.Algorithms.If(contained, init, init.add(element))
    return ee.List(eelist.iterate(wrap, newlist))


def removeIndex(list, index):
    """ Remove an element by its index """
    list = ee.List(list)
    index = ee.Number(index)
    size = list.size()

    def allowed():
        def zerof(list):
            return list.slice(1, list.size())

        def rest(list, index):
            list = ee.List(list)
            index = ee.Number(index)
            last = index.eq(list.size())

            def lastf(list):
                return list.slice(0, list.size().subtract(1))

            def restf(list, index):
                list = ee.List(list)
                index = ee.Number(index)
                first = list.slice(0, index)
                return first.cat(list.slice(index.add(1), list.size()))

            return ee.List(ee.Algorithms.If(last, lastf(list), restf(list, index)))

        return ee.List(ee.Algorithms.If(index, rest(list, index), zerof(list)))

    condition = index.gte(size).Or(index.lt(0))

    return ee.List(ee.Algorithms.If(condition, -1, allowed()))


def replaceDict(eelist, to_replace):
    """ Replace many elements of a Earth Engine List object using a dictionary

        **EXAMPLE**

    .. code:: python

        list = ee.List(["one", "two", "three", 4])
        newlist = replace_many(list, {"one": 1, 4:"four"})

        print newlist.getInfo()

    >> [1, "two", "three", "four"]

    :param to_replace: values to replace
    :type to_replace: dict
    :return: list with replaced values
    :rtype: ee.List
    """
    eelist = ee.List(eelist)
    to_replace = ee.Dictionary(to_replace)
    keys = to_replace.keys()
    def wrap(el):
        # Convert to String
        elstr = ee.Algorithms.String(el)
        condition = ee.List(keys).indexOf(elstr)
        return ee.Algorithms.If(condition.neq(-1), to_replace.get(elstr), el)
    return eelist.map(wrap)


def sequence(ini, end, step=1):
    """ Create a sequence from ini to end by step. Similar to
    ee.List.sequence, but if end != last item then adds the end to the end
    of the resuting list
    """
    end = ee.Number(end)
    if step == 0:
        step = 1
    amplitude = end.subtract(ini)
    mod = ee.Number(amplitude).mod(step)
    seq = ee.List.sequence(ini, end, step)
    condition = mod.neq(0)
    final = ee.Algorithms.If(condition, seq.add(end), seq)
    return ee.List(final)


def toString(eelist):
    """ Convert elements of a list into Strings. If the list contains other
    elements that are not strings or numbers, it will return the object type.
    For example, ['a', 1, ee.Image(0)] -> ['a', '1', 'Image']
    """
    eelist = ee.List(eelist)
    def wrap(el):
        def false(el):
            otype = ee.Algorithms.ObjectType(el)
            return ee.String(ee.Algorithms.If(computedobject.isNumber(el),
                                              ee.Number(el).format(),
                                              otype))

        return ee.String(ee.Algorithms.If(computedobject.isString(el), el, false(el)))

    return eelist.map(wrap)


def zip(eelist):
    """ Zip a list of lists.

    Example:
        nested = ee.List([[1,2,3], [4,5,6], [7,8,9]])
        zipped = geetools.tools.ee_list.zip(nested)
        print(zipped.getInfo())
        >> [[1, 4, 7], [2, 5, 8], [3, 6, 9]]
    """
    eelist = ee.List(eelist)
    first = ee.List(eelist.get(0))
    rest = ee.List(eelist).slice(1)
    def wrap(l, accum):
        accum = ee.List(accum)
        return accum.zip(l).map(lambda l: ee.List(l).flatten())
    return ee.List(rest.iterate(wrap, first))


# coding=utf-8
import ee
import os
from . import utils
from ..utils import makeName
from .. import tools


def toDrive(collection, folder, namePattern='{id}', scale=30,
            dataType="float", region=None, datePattern=None,
            extra=None, verbose=False, **kwargs):
    """ Upload all images from one collection to Google Drive. You can use
    the same arguments as the original function
    ee.batch.export.image.toDrive

    :param collection: Collection to upload
    :type collection: ee.ImageCollection
    :param folder: Google Drive folder to export the images to
    :type folder: str
    :param namePattern: pattern for the name.
        See geetools.tools.image.make_name function and also
        geetools.tools.string.format function
    :type namePattern: str
    :param scale: scale of the image (side of one pixel). Defults to 30
        (Landsat resolution)
    :type scale: int
    :param dataType: as downloaded images **must** have the same data type
        in all bands, you have to set it here. Can be one of: "float",
        "double", "int", "Uint8", "Int8" or a casting function like
        *ee.Image.toFloat*
    :type dataType: str
    :param region: area to upload. Default to the footprint of the first
        image in the collection
    :type region: ee.Geometry.Rectangle or ee.Feature
    :param datePattern: pattern for date if specified in namePattern.
        Defaults to 'yyyyMMdd'. See ee.Date.format for more details
    :type datePattern: str
    :param extra: extra parameters to parse to the name formatter
    :type extra: str
    :param verbose: print name of each exporting task
    :type verbose: bool
    :return: list of tasks
    :rtype: list
    """
    # empty tasks list
    tasklist = []
    # get region
    if region:
        region = tools.geometry.getRegion(region)

    # Make a list of images
    img_list = collection.toList(collection.size())

    n = 0
    while True:
        try:
            img = ee.Image(img_list.get(n))

            name = makeName(img, namePattern, datePattern, extra)
            name = name.getInfo()
            description = utils.matchDescription(name)

            # convert data type
            img = utils.convertDataType(dataType)(img)

            if region is None:
                region = tools.geometry.getRegion(img)

            task = ee.batch.Export.image.toDrive(image=img,
                                                 description=description,
                                                 folder=folder,
                                                 fileNamePrefix=name,
                                                 region=region,
                                                 scale=scale, **kwargs)
            task.start()
            if verbose:
                print("exporting {} to folder '{}' in GDrive".format(name, folder))

            tasklist.append(task)
            n += 1
        except Exception as e:
            error = str(e).split(':')
            if error[0] == 'List.get':
                break
            else:
                raise e

    return tasklist


def toCloudStorage(collection, bucket, folder=None, namePattern='{id}',
                   region=None, scale=30, dataType="float", datePattern=None,
                   verbose=False, extra=None, **kwargs):
    """ Upload all images from one collection to Google Cloud Storage. You can
    use the same arguments as the original function
    ee.batch.export.image.toCloudStorage

    :param collection: Collection to upload
    :type collection: ee.ImageCollection
    :param bucket: Google Cloud Storage bucket name
    :type folder: str
    :param folder: Google Cloud Storage prefix to export the images to
    :type folder: str
    :param namePattern: pattern for the name. See make_name function
    :type namePattern: str
    :param region: area to upload. Defualt to the footprint of the first
        image in the collection
    :type region: ee.Geometry.Rectangle or ee.Feature
    :param scale: scale of the image (side of one pixel). Defults to 30
        (Landsat resolution)
    :type scale: int
    :param dataType: as downloaded images **must** have the same data type
        in all bands, you have to set it here. Can be one of: "float",
        "double", "int", "Uint8", "Int8" or a casting function like
        *ee.Image.toFloat*
    :type dataType: str
    :param datePattern: pattern for date if specified in namePattern.
        Defaults to 'yyyyMMdd'
    :type datePattern: str
    """
    # empty tasks list
    tasklist = []
    # get region
    region = tools.geometry.getRegion(region)
    # Make a list of images
    img_list = collection.toList(collection.size())

    n = 0
    while True:
        try:
            img = ee.Image(img_list.get(n))

            name = makeName(img, namePattern, datePattern, extra)
            name = name.getInfo()
            description = utils.matchDescription(name)

            # convert data type
            img = utils.convertDataType(dataType)(img)

            if folder is not None:
                path = folder + "/" + name
            else:
                path = name

            task = ee.batch.Export.image.toCloudStorage(image=img,
                                                 description=description,
                                                 bucket=bucket,
                                                 path=path,
                                                 region=region,
                                                 scale=scale,
                                                 **kwargs)
            task.start()
            tasklist.append(task)
            if verbose:
                print("adding task {} to list".format(name))
            n += 1

        except Exception as e:
            error = str(e).split(':')
            if error[0] == 'List.get':
                break
            else:
                raise e


def toAsset(collection, assetPath, namePattern=None, scale=30, region=None,
            create=True, verbose=False, datePattern='yyyyMMdd',
            extra=None, **kwargs):
    """ Upload all images from one collection to a Earth Engine Asset.
    You can use the same arguments as the original function
    ee.batch.export.image.toDrive

    :param collection: Collection to upload
    :type collection: ee.ImageCollection
    :param assetPath: path of the asset where images will go
    :type assetPath: str
    :param namePattern: pattern for the name. If None, it uses the posion
        of the image in the image collection as the name. Otherwise see
        geetools.tools.image.make_name function and also
        geetools.tools.string.format function
    :type namePattern: str
    :param region: area to upload. Defualt to the footprint of the first
        image in the collection
    :type region: ee.Geometry.Rectangle or ee.Feature
    :param scale: scale of the image (side of one pixel). Defults to 30
        (Landsat resolution)
    :type scale: int
    :return: list of tasks
    :rtype: list
    """
    tasklist = []

    if create:
        utils.createAssets([assetPath], 'ImageCollection', True)

    if region:
        region = tools.geometry.getRegion(region)

    if namePattern:
        imlist = collection.toList(collection.size())
        idx = 0
        while True:
            try:
                img = imlist.get(idx)
                img = ee.Image(img)
                name = makeName(img, namePattern, datePattern, extra)
                name = name.getInfo()
                description = utils.matchDescription(name)
                assetId = assetPath+"/"+name

                if region is None:
                    region = tools.geometry.getRegion(img)

                task = ee.batch.Export.image.toAsset(image=img,
                                                     assetId=assetId,
                                                     description=description,
                                                     region=region,
                                                     scale=scale, **kwargs)
                task.start()

                if verbose:
                    print('Exporting {} to {}'.format(name, assetId))

                tasklist.append(task)
                idx += 1
            except Exception as e:
                error = str(e).split(':')
                if error[0] == 'List.get':
                    break
                else:
                    raise e

    else:
        size = collection.size().getInfo()
        imlist = collection.toList(size)
        for idx in range(0, size+1):
            img = imlist.get(idx)
            img = ee.Image(img)
            name = str(idx)
            description = name
            assetId = assetPath+"/"+name

            if region is None:
                region = tools.geometry.getRegion(img)

            task = ee.batch.Export.image.toAsset(image=img,
                                                 assetId=assetId,
                                                 description=description,
                                                 region=region,
                                                 scale=scale, **kwargs)
            task.start()

            if verbose:
                print('Exporting {} to {}'.format(name, assetId))

            tasklist.append(task)

    return tasklist


def qgisCode(collection, visParams=None, name=None, datePattern=None,
             verbose=False):
    QGIS_COL_CODE = """names={names}
urls={urls}
for name, url in zip(names, urls):
    urlWithParams = "type=xyz&url={{}}".format(url)
    rlayer = QgsRasterLayer(urlWithParams, name, "wms")
    if rlayer.isValid():
        QgsProject.instance().addMapLayer(rlayer)
    else:
        print("invalid layer")
"""
    name = name or '{id}'
    names = []
    urls = []
    i = 0
    collist = collection.toList(collection.size())
    catch = 'List.get: List index must be between'
    while True:
        try:
            img = ee.Image(collist.get(i))
            n = makeName(img, name, datePattern).getInfo()
            if verbose:
                print('processing {}'.format(n))
            url = tools.image.getTileURL(img, visParams)
            names.append(n)
            urls.append(url)
            i += 1
        except Exception as e:
            if catch in str(e):
                break
            else:
                raise e

    return QGIS_COL_CODE.format(names=names, urls=urls)


def toQGIS(collection, visParams=None, name=None, filename=None, path=None,
           datePattern=None, replace=True, verbose=True):
    """ Download a python file to import from QGIS """
    code = qgisCode(collection, visParams, name, datePattern, verbose)
    path = path or os.getcwd()
    # Check extension
    if filename:
        ext = filename.split('.')[-1]
        if ext != 'py':
            filename += '.py'
    else:
        filename = 'qgis2ee'
    # add _qgis_ to filename
    splitted = filename.split('.')[:-1]
    noext = '.'.join(splitted)
    filename = '{}_qgis_'.format(noext)
    # process
    finalpath = os.path.join(path, filename)
    finalpath = '{}.py'.format(finalpath)
    if not os.path.exists(finalpath) or replace:
        with open(finalpath, 'w+') as thefile:
            thefile.write(code)
        return thefile
    else:
        return None



# coding=utf-8

import ee
from geetools import cloud_mask
from geetools.tools.image import getValue
from . import TEST_CLOUD_IMAGES

# Initialize
ee.Initialize()

p_cloud = ee.Geometry.Point([-69.7850919885049, -42.74847761369703])
p_snow = ee.Geometry.Point([-69.95680141007878, -43.08818524620294])
p_shadow = ee.Geometry.Point([-69.85037130586764, -42.78657084334236])

image = ee.Image(TEST_CLOUD_IMAGES['L5TOA'])


def test_clouds():
    masked = cloud_mask.landsat457ToaBQA(['cloud'])(image)
    vals = getValue(masked, p_cloud, 30)

    assert vals.get("B1").getInfo() == None


def test_shadows():
    masked = cloud_mask.landsat457ToaBQA(['shadow'])(image)
    vals = getValue(masked, p_shadow, 30)

    assert vals.get("B1").getInfo() == None


def test_snow():
    masked = cloud_mask.landsat457ToaBQA(['snow'])(image)
    vals = getValue(masked, p_snow, 30)

    assert vals.get("B1").getInfo() == None


# coding=utf-8
"""
Set of custom filters for Google Earth Engine.
"""
from __future__ import print_function
import ee


def dateRange(range):
    """ Filter by DateRange

    :param range: date range
    :type range: ee.DateRange
    :return: the filter to apply to a collection
    :rtype: ee.Filter
    """
    ini = range.start()
    end = range.end()
    return ee.Filter.date(ini, end)


# coding=utf-8

from . import featurecollection, imagecollection, utils
from . import image as batchimage
import ee


class Export(ee.batch.Export):
    class table(ee.batch.Export.table):
        toAsset = featurecollection.toAsset

    class imagecollection(object):
        toDrive = imagecollection.toDrive
        toAsset = imagecollection.toAsset
        toCloudStorage = imagecollection.toCloudStorage

    class image(ee.batch.Export.image):
        toAsset = batchimage.toAsset  # overwrite original
        toDriveByFeature = batchimage.toDriveByFeature


class Download(object):
    class table(object):
        toGeoJSON = featurecollection.toGeoJSON
        toCSV = featurecollection.toCSV
        toLocal = featurecollection.toLocal

    class imagecollection(object):
        toQGIS = imagecollection.toQGIS

    class image(object):
        toLocal = batchimage.toLocal
        toQGIS = batchimage.toQGIS


class Convert(object):
    class table(object):
        toDict = featurecollection.toDict

    class imagecollection(object):
        pass

    class image(object):
        pass


class Import(object):
    class table(object):
        fromGeoJSON = featurecollection.fromGeoJSON
        fromShapefile = featurecollection.fromShapefile
        fromKML = featurecollection.fromKML

    class imagecollection(object):
        pass

    class image(object):
        pass


# coding=utf-8

import ee
from geetools import cloud_mask
from geetools.tools.image import getValue
from . import TEST_CLOUD_IMAGES

# Initialize
ee.Initialize()

p_cloud = ee.Geometry.Point([-71.67030933850714, -42.92352111915706])
p_snow = ee.Geometry.Point([-71.75856720476548, -42.93871143148904])
p_shadow = ee.Geometry.Point([-71.67144910085082, -43.0356524986155])

image = ee.Image(TEST_CLOUD_IMAGES['L4TOA'])


def test_clouds():
    masked = cloud_mask.landsat457ToaBQA(['cloud'])(image)
    vals = getValue(masked, p_cloud, 30)

    assert vals.get("B1").getInfo() == None


def test_shadows():
    masked = cloud_mask.landsat457ToaBQA(['shadow'])(image)
    vals = getValue(masked, p_shadow, 30)

    assert vals.get("B1").getInfo() == None


def test_snow():
    masked = cloud_mask.landsat457ToaBQA(['snow'])(image)
    vals = getValue(masked, p_snow, 30)

    assert vals.get("B1").getInfo() == None


# coding=utf-8
""" Helper functions for visualizing """
import ee


def stretch_std(image, region, bands=None, std=1, scale=None):
    """ Get mins and maxs values for stretching a visualization using standard
    deviation """
    if not bands:
        names = image.bandNames()
        bands = ee.List(ee.Algorithms.If(
            names.size().gte(3), names.slice(0,3), names.slice(0)))
        bands = bands.getInfo()

    image = image.select(bands)
    geom = region or image.geometry()
    params = dict(geometry=geom, bestEffort=True)
    if scale:
        params['scale'] = scale

    params['reducer'] = ee.Reducer.mean()
    mean = image.reduceRegion(**params)

    params['reducer'] = ee.Reducer.stdDev()
    stdDev = image.reduceRegion(**params)

    def minmax(band, val):
        minv = ee.Number(val).subtract(ee.Number(stdDev.get(band)).multiply(std))
        maxv = ee.Number(val).add(ee.Number(stdDev.get(band)).multiply(std))
        return ee.List([minv, maxv])

    if len(bands) == 1:
        band = bands[0]
        values = minmax(band, mean.get(band)).getInfo()
        minv = values[0]
        maxv = values[1]
    else:
        values = mean.map(minmax).select(bands).getInfo()
        minv = [values[bands[0]][0], values[bands[1]][0], values[bands[2]][0]]
        maxv = [values[bands[0]][1], values[bands[1]][1], values[bands[2]][1]]

    return dict(bands=bands, min=minv, max=maxv)



def stretch_percentile(image, region, bands=None, percentile=90, scale=None):
    """ Get mins and maxs values for stretching a visualization using
    percentiles """
    # Calculate start and end percentiles
    startp = 50-(percentile/2)
    endp = 50+(percentile/2)

    if not bands:
        names = image.bandNames()
        bands = ee.List(ee.Algorithms.If(
            names.size().gte(3), names.slice(0,3), names.slice(0)))
        bands = bands.getInfo()

    image = image.select(bands)
    geom = region or image.geometry()
    params = dict(geometry=geom, bestEffort=True)
    if scale:
        params['scale'] = scale

    params['reducer'] = ee.Reducer.percentile([startp, endp])
    percentiles = image.reduceRegion(**params)

    def minmax(band):
        minkey = ee.String(band).cat('_p').cat(ee.Number(startp).format())
        maxkey = ee.String(band).cat('_p').cat(ee.Number(endp).format())

        minv = ee.Number(percentiles.get(minkey))
        maxv = ee.Number(percentiles.get(maxkey))
        return ee.List([minv, maxv])

    if len(bands) == 1:
        band = bands[0]
        values = minmax(band).getInfo()
        minv = values[0]
        maxv = values[1]
    else:
        values = ee.List(bands).map(minmax).getInfo()
        minv = [values[0][0], values[1][0], values[2][0]]
        maxv = [values[0][1], values[1][1], values[2][1]]

    return dict(bands=bands, min=minv, max=maxv)



# coding=utf-8

import ee
from geetools import cloud_mask
from geetools.tools.image import getValue
from . import TEST_CLOUD_IMAGES

# Initialize
ee.Initialize()

p_cloud = ee.Geometry.Point([-70.15588363762427, -42.98086920709981])
p_snow = ee.Geometry.Point([-69.9144023012539, -42.73663097771261])
p_shadow = ee.Geometry.Point([-70.3042273263779, -42.70884337439131])

image = ee.Image(TEST_CLOUD_IMAGES['L8SR'])


def test_clouds():
    masked = cloud_mask.landsat8SRPixelQA(['cloud'])(image)
    vals = getValue(masked, p_cloud, 30)

    assert vals.get("B1").getInfo() == None


def test_shadows():
    masked = cloud_mask.landsat8SRPixelQA(['shadow'])(image)
    vals = getValue(masked, p_shadow, 30)

    assert vals.get("B1").getInfo() == None


def test_snow():
    masked = cloud_mask.landsat8SRPixelQA(['snow'])(image)
    vals = getValue(masked, p_snow, 30)

    assert vals.get("B1").getInfo() == None


# coding=utf-8

import ee
from geetools import cloud_mask
from geetools.tools.image import getValue
from . import TEST_CLOUD_IMAGES

# Initialize
ee.Initialize()

p_cloud = ee.Geometry.Point([-70.47305608720774, -43.23185291813854])
p_snow = ee.Geometry.Point([-71.45790258078314, -42.65855880983149])
p_shadow = ee.Geometry.Point([-70.80613567662904, -43.01702683734614])

image = ee.Image(TEST_CLOUD_IMAGES['L5SR'])


def test_clouds():
    masked = cloud_mask.landsat457SRPixelQA(['cloud'])(image)
    vals = getValue(masked, p_cloud, 30)

    assert vals.get("B1").getInfo() == None


def test_shadows():
    masked = cloud_mask.landsat457SRPixelQA(['shadow'])(image)
    vals = getValue(masked, p_shadow, 30)

    assert vals.get("B1").getInfo() == None


def test_snow():
    masked = cloud_mask.landsat457SRPixelQA(['snow'])(image)
    vals = getValue(masked, p_snow, 30)

    assert vals.get("B1").getInfo() == None


# coding=utf-8
""" Module holding misc algorithms """
import ee
import ee.data
import math
from . import tools


def distanceToMask(image, kernel=None, radius=1000, unit='meters',
                   scale=None, geometry=None, band_name='distance_to_mask',
                   normalize=False):
    """ Compute the distance to the mask in meters

    :param image: Image holding the mask
    :type image: ee.Image
    :param kernel: Kernel to use for computing the distance. By default uses
        euclidean
    :type kernel: ee.Kernel
    :param radius: radius for the kernel. Defaults to 1000
    :type radius: int
    :param unit: units for the kernel radius. Defaults to 'meters'
    :type unit: str
    :param scale: scale for reprojection. If None, will reproject on the fly
        (according to EE lazy computing)
    :type scale: int
    :param geometry: compute the distance only inside this geometry. If you
        want to compute the distance inside a clipped image, using this
        parameter will make the edges not be considered as a mask.
    :type geometry: ee.Geometry or ee.Feature
    :param band_name: name of the resulting band. Defaults to
        'distance_to_mask'
    :type band_name: str
    :param normalize: Normalize result (between 0 and 1)
    :type normalize: bool
    :return: A one band image with the distance to the mask
    :rtype: ee.Image
    """
    if not kernel:
        kernel = ee.Kernel.euclidean(radius, unit)

    # select first band
    image = image.select(0)

    # get mask
    mask = image.mask()
    inverse = mask.Not()

    if geometry:
        # manage geometry types
        if isinstance(geometry, (ee.Feature, ee.FeatureCollection)):
            geometry = geometry.geometry()

        inverse = inverse.clip(geometry)

    # Compute distance to the mask (inverse)
    distance = inverse.distance(kernel)

    if scale:
        proj = image.projection()
        distance = distance.reproject(proj.atScale(scale))

    # make mask to be the max distance
    dist_mask = distance.mask().Not().remap([0, 1], [0, radius])

    if geometry:
        dist_mask = dist_mask.clip(geometry)

    final = distance.unmask().add(dist_mask)

    if normalize:
        final = tools.image.parametrize(final, (0, radius), (0, 1))

    return final.rename(band_name)


def maskCover(image, geometry=None, scale=None, property_name='MASK_COVER',
              crs=None, crsTransform=None, bestEffort=False,
              maxPixels=1e13, tileScale=1):
    """ Percentage of masked pixels (masked/total * 100) as an Image property

    :param image: ee.Image holding the mask. If the image has more than
        one band, the first one will be used
    :type image: ee.Image
    :param geometry: the value will be computed inside this geometry. If None,
        will use image boundaries. If unbounded the result will be 0
    :type geometry: ee.Geometry or ee.Feature
    :param scale: the scale of the mask
    :type scale: int
    :param property_name: the name of the resulting property
    :type property_name: str
    :return: The same parsed image with a new property holding the mask cover
        percentage
    :rtype: ee.Image
    """
    # keep only first band
    imageband = image.select(0)

    # get projection
    projection = imageband.projection()

    if not scale:
        scale = projection.nominalScale()

    # get band name
    band = ee.String(imageband.bandNames().get(0))

    # Make an image with all ones
    ones_i = ee.Image.constant(1).reproject(projection).rename(band)

    if not geometry:
        geometry = image.geometry()

    # manage geometry types
    if isinstance(geometry, (ee.Feature, ee.FeatureCollection)):
        geometry = geometry.geometry()

    unbounded = geometry.isUnbounded()

    # Get total number of pixels
    ones = ones_i.reduceRegion(
        reducer= ee.Reducer.count(),
        geometry= geometry,
        scale= scale,
        maxPixels= maxPixels,
        crs=crs,
        crsTransform=crsTransform,
        bestEffort=bestEffort,
        tileScale=tileScale,
    ).get(band)
    ones = ee.Number(ones)

    # select first band, unmask and get the inverse
    mask = imageband.mask()
    mask_not = mask.Not()
    image_to_compute = mask.updateMask(mask_not)

    # Get number of zeros in the given image
    zeros_in_mask =  image_to_compute.reduceRegion(
        reducer= ee.Reducer.count(),
        geometry= geometry,
        scale= scale,
        maxPixels= maxPixels,
        crs=crs,
        crsTransform=crsTransform,
        bestEffort=bestEffort,
        tileScale=tileScale
    ).get(band)
    zeros_in_mask = ee.Number(zeros_in_mask)

    percentage = tools.number.trimDecimals(zeros_in_mask.divide(ones), 4)

    # Multiply by 100
    cover = percentage.multiply(100)

    # Return None if geometry is unbounded
    final = ee.Number(ee.Algorithms.If(unbounded, 0, cover))

    return image.set(property_name, final)


def euclideanDistance(image1, image2, bands=None, discard_zeros=False,
                      name='distance'):
    """ Compute the Euclidean distance between two images. The image's bands
    is the dimension of the arrays.

    :param image1:
    :type image1: ee.Image
    :param image2:
    :type image2: ee.Image
    :param bands: the bands that want to be computed
    :type bands: list
    :param discard_zeros: pixel values equal to zero will not count in the
        distance computation
    :type discard_zeros: bool
    :param name: the name of the resulting band
    :type name: str
    :return: a distance image
    :rtype: ee.Image
    """
    if not bands:
        bands = image1.bandNames()

    image1 = image1.select(bands)
    image2 = image2.select(bands)

    proxy = tools.image.empty(0, bands)
    image1 = proxy.where(image1.gt(0), image1)
    image2 = proxy.where(image2.gt(0), image2)

    if discard_zeros:
        # zeros
        zeros1 = image1.eq(0)
        zeros2 = image2.eq(0)

        # fill zeros with values from the other image
        image1 = image1.where(zeros1, image2)
        image2 = image2.where(zeros2, image1)

    a = image1.subtract(image2)
    b = a.pow(2)
    c = b.reduce('sum')
    d = c.sqrt()

    return d.rename(name)


def sumDistance(image, collection, bands=None, discard_zeros=False,
                name='sumdist'):
    """ Compute de sum of all distances between the given image and the
    collection passed
    
    :param image: 
    :param collection:
    :return: 
    """
    condition = isinstance(collection, ee.ImageCollection)

    if condition:
        collection = collection.toList(collection.size())

    accum = ee.Image(0).rename(name)

    def over_rest(im, ini):
        ini = ee.Image(ini)
        im = ee.Image(im)
        dist = ee.Image(euclideanDistance(image, im, bands, discard_zeros))\
                 .rename(name)
        return ini.add(dist)

    return ee.Image(collection.iterate(over_rest, accum))


def pansharpenKernel(image, pan, rgb=None, kernel=None):
    """
    Compute the per-pixel means of the unsharpened bands
    source: https://gis.stackexchange.com/questions/296615/pansharpen-landsat-mosaic-in-google-earth-engine

    :param pan: the name of the panchromatic band
    :type pan: str
    :param rgb: the red green blue bands
    :type rgb: tuple or list
    :param kernel: the kernel to reduce neighbors
    :type kernel: ee.Kernel
    :rtype: ee.Image
    """
    if not kernel:
        kernel = ee.Kernel.square(90, 'meters')
    if not rgb:
        rgb = ['red', 'green', 'blue']

    if not pan:
        pan = 'pan'

    bgr = image.select(rgb)
    pani = image.select(pan)
    bgr_mean = bgr.reduce('mean').rename('mean')
    # Compute the aggregate mean of the unsharpened bands and the pan band
    mean_values = pani.addBands(bgr_mean).reduceNeighborhood(
        ee.Reducer.mean(),
        kernel
    )
    gain = mean_values.select('mean_mean').divide(
        mean_values.select('{}_mean'.format(pan)))
    sharpen = bgr.divide(bgr_mean).multiply(pani).multiply(gain)
    return ee.Image(sharpen.copyProperties(source=image).set(
        'system:time_start', image.date().millis()))


def pansharpenIhsFusion(image, pan=None, rgb=None):
    """
    HSV-based Pan-Sharpening
    source: https://gis.stackexchange.com/questions/296615/pansharpen-landsat-mosaic-in-google-earth-engine

    :param image:
    :type image: ee.Image
    the name of the panchromatic band
    :type pan: str
    :param rgb: the red green blue bands
    :type rgb: tuple or list
    :rtype: ee.Image
    """
    if not rgb:
        rgb = ['red', 'green', 'blue']

    if not pan:
        pan = 'pan'

    rgb = image.select(rgb)
    pan = image.select(pan)
    # Convert to HSV, swap in the pan band, and convert back to RGB.
    huesat = rgb.rgbToHsv().select('hue', 'saturation')
    upres = ee.Image.cat(huesat, pan).hsvToRgb()
    return image.addBands(upres)


class Landsat(object):

    @staticmethod
    def unmask_slc_off(image, optical_bands='B.+'):
        """ Unmask pixels that were affected by scl-off  error in Landsat 7

        Expects a Landsat 7 image and it is meant to be used before any other
        masking, otherwise this could affect the previous mask.

        The parameter `optical_bands` must be used if band were renamed. By
        default it is `B.+` which means that optical bands are those starting
        with B: B1 (blue), B2 (green), B3 (red), B4 (nir), B5 (swir),
        B6 (thermal), B7 (swir2).
        """
        idate = image.date()
        slcoff = ee.Date('2003-05-31')
        condition = idate.difference(slcoff, 'days').gte(0)

        def compute(i):
            mask = i.mask()
            reduced = mask.select(optical_bands).reduce('sum')
            slc_off = reduced.eq(0)
            unmasked = i.unmask()
            newmask = mask.where(slc_off, 1)
            return unmasked.updateMask(newmask)

        return ee.Image(ee.Algorithms.If(condition, compute(image), image))


    @staticmethod
    def _rescale(image, bands=None, thermal_bands=None, original='TOA',
                 to='SR', number='all'):
        """ Rescaling logic """
        if not bands:
            bands = ['B1','B2','B3','B4','B5','B6','B7']
        bands = ee.List(bands)

        if not thermal_bands:
            thermal_bands = ['B10', 'B11']
        thermal_bands = ee.List(thermal_bands)

        allbands = bands.cat(thermal_bands)

        if number == '8':
            max_raw = 65535
        else:
            max_raw = 255

        if original == 'TOA' and to == 'SR':
            scaled = image.select(bands).multiply(10000).toInt16()
            scaled_thermal = image.select(thermal_bands).multiply(10).toInt16()
        elif original == 'SR' and to == 'TOA':
            scaled = image.select(bands).toFloat().divide(10000)
            scaled_thermal = image.select(thermal_bands).toFloat().divide(10)
        elif original == 'TOA' and to == 'RAW':
            scaled = tools.image.parametrize(image.select(bands), (0, 1),
                                             (0, max_raw))
            scaled_thermal = tools.image.parametrize(image.select(bands), (0, 1),
                                                     (0, max_raw)).multiply(1000)

        original_bands = image.bandNames()

        rest_bands = tools.ee_list.difference(original_bands, allbands)

        rest_image = image.select(rest_bands)

        return rest_image.addBands(scaled).addBands(scaled_thermal)

    @staticmethod
    def rescaleToaSr(image, bands=None, thermal_bands=None):
        """ Re-scale a TOA Landsat image to match the data type of SR Landsat
        image

        :param image: a Landsat TOA image
        :type image: ee.Image
        :param bands: bands to rescale by 10000. If None will use the bands for
            Landsat 8 (['B1','B2','B3','B4','B5','B6','B7']).
        :type bands: list
        :param thermal_bands: bands to rescale by 100. Defaults to
            ['B10', 'B11']
        :type thermal_bands: list
        :rtype: ee.Image
        """
        return Landsat._rescale(image, bands, thermal_bands, 'TOA', 'SR')

    @staticmethod
    def rescaleSrToa(image, bands=None, thermal_bands=None):
        """ Re-scale a TOA Landsat image to match the data type of SR Landsat
        image

        :param image: a Landsat TOA image
        :type image: ee.Image
        :param bands: bands to rescale by 10000. If None will use the bands for
            Landsat 8 (B[1-7]).
        :type bands: list
        :param thermal_bands: bands to rescale by 100. Defaults to
            ['B10', 'B11']
        :type thermal_bands: list
        :rtype: ee.Image
        """
        return Landsat._rescale(image, bands, thermal_bands, 'SR', 'TOA')

    @staticmethod
    def harmonization(image, blue='B2', green='B3', red='B4', nir='B5',
                      swir='B6', swir2='B7', max_value=None):
        """ Harmonization of Landsat 8 images to be consistant with
        Landsat 7 images

        Roy, D.P., Kovalskyy, V., Zhang, H.K., Vermote, E.F., Yan, L.,
        Kumar, S.S, Egorov, A., 2016, Characterization of Landsat-7 to
        Landsat-8 reflective wavelength and normalized difference vegetation
        index continuity, Remote Sensing of Environment, 185, 57-70.
        (http://dx.doi.org/10.1016/j.rse.2015.12.024) Table 2 -
        reduced major axis (RMA) regression coefficients

        :param image: A Landsat 8 Image
        :param max_value: the maximum value for the optical bands. For float
            bands it is 1 (TOA), for int16 it is 10000 (SR) and for int8 it is
            255 (RAW). It default to 1 (TOA)
        :return:
        """
        bands = ee.List([blue, green, red, nir, swir, swir2])
        band_types = image.bandTypes().select(bands)

        if max_value is None:
            max_value = 1

        slopes = ee.Image.constant([0.9785, 0.9542, 0.9825,
                                    1.0073, 1.0171, 0.9949])
        itcp = ee.Image.constant([-0.0095, -0.0016, -0.0022,
                                  -0.0021, -0.0030, 0.0029])

        only_bands = image.select(bands)
        resampled = only_bands.resample('bicubic')

        harmonized = resampled.subtract(itcp.multiply(max_value)).divide(slopes)

        harmonized = harmonized.cast(band_types)

        # append rest of the bands
        return image.addBands(harmonized, overwrite=True)

    @staticmethod
    def brdfCorrect(image, red='red', green='green', blue='blue', nir='nir',
                    swir1='swir1', swir2='swir2'):
        """ Correct Landsat data for BRDF effects using a c-factor.

        D.P. Roy, H.K. Zhang, J. Ju, J.L. Gomez-Dans, P.E. Lewis, C.B. Schaaf,
        Q. Sun, J. Li, H. Huang, V. Kovalskyy, A general method to normalize
        Landsat reflectance data to nadir BRDF adjusted reflectance,
        Remote Sensing of Environment, Volume 176, April 2016, Pages 255-271

        Interpreted and coded here by Daniel Wiell and Erik Lindquist of the
        United Nations Food and Agriculture Organization

        Transcipted to GEE Python API by Rodrigo E. Principe

        If the band names of the passed image are 'blue', 'green', etc,
        those will be used, if not, relations must be indicated in params.

        :rtype: ee.Image
        """
        original = image
        constants = {'pi': math.pi}

        ### HELPERS ###
        def merge(o1, o2):
            def addAll(target, toAdd):
                for key, val in toAdd.items():
                    target[key] = val

            result = {}
            addAll(result, o1)
            addAll(result, o2)
            return result

        def format_str(string, args=None):
            args = args if args else {}
            allArgs = merge(constants, args)
            return string.format(**allArgs)

        def toImage(img, band, args=None):
            """
            compute an expression passed in band if it's a str.
            formats the expression in band using format_str if necessary
            :return: one band image
            """
            if isinstance(band, str):
                # print('band:', band)
                if (band.find('.') > -1) \
                        or (band.find(' ') > -1) \
                        or (band.find('{') > -1):
                    # print('formatted:', format_str(band, args), '\n')
                    band = img.expression(format_str(band, args), {'i': img})
                else:
                    band = image.select(band)

            return ee.Image(band)

        def set_name(img, name, toAdd, args=None):
            """
            compute the band (toAdd) with toImage.
            add the band to the passed image and rename it
            """
            toAdd = toImage(img, toAdd, args)
            return img.addBands(toAdd.rename(name), None, True)

        def setIf(img, name, condition=None, trueValue=None, falseValue=None):

            def invertMask(mask):
                # return mask.multiply(-1).add(1)
                return mask.Not()

            condition = toImage(img, condition)
            trueMasked = toImage(img, trueValue).mask(toImage(img, condition))
            falseMasked = toImage(img, falseValue).mask(invertMask(condition))
            value = trueMasked.unmask(falseMasked)
            return set_name(img, name, value)

        def x(point):
            return ee.Number(ee.List(point).get(0))

        def y(point):
            return ee.Number(ee.List(point).get(1))

        def pointBetween(pointA, pointB):
            return ee.Geometry.LineString([pointA, pointB]).centroid() \
                .coordinates()

        def slopeBetween(pointA, pointB):
            return ((y(pointA)).subtract(y(pointB))) \
                .divide((x(pointA)).subtract(x(pointB)))

        def toLine(pointA, pointB):
            return ee.Geometry.LineString([pointA, pointB])

        ### END HELPERS ###

        # image = image.select([red, green, blue, nir, swir1, swir2],
        #                      ['red', 'green', 'blue', 'nir', 'swir1', 'swir2'])

        inputBandNames = image.bandNames()

        coefficientsByBand = {
            blue: {'fiso': 0.0774, 'fgeo': 0.0079, 'fvol': 0.0372},
            green: {'fiso': 0.1306, 'fgeo': 0.0178, 'fvol': 0.0580},
            red: {'fiso': 0.1690, 'fgeo': 0.0227, 'fvol': 0.0574},
            nir: {'fiso': 0.3093, 'fgeo': 0.0330, 'fvol': 0.1535},
            swir1: {'fiso': 0.3430, 'fgeo': 0.0453, 'fvol': 0.1154},
            swir2: {'fiso': 0.2658, 'fgeo': 0.0387, 'fvol': 0.0639}
        }

        def findCorners(img):
            footprint = ee.Geometry(img.get('system:footprint'))
            bounds = ee.List(footprint.bounds().coordinates().get(0))
            coords = footprint.coordinates()

            def wrap_xs(item):
                return x(item)
            xs = coords.map(wrap_xs)

            def wrap_ys(item):
                return y(item)

            ys = coords.map(wrap_ys)

            def findCorner(targetValue, values):
                def wrap_diff(value):
                    return ee.Number(value).subtract(targetValue).abs()
                diff = values.map(wrap_diff)
                minValue = diff.reduce(ee.Reducer.min())
                idx = diff.indexOf(minValue)
                return ee.Number(coords.get(idx))

            lowerLeft = findCorner(x(bounds.get(0)), xs)
            lowerRight = findCorner(y(bounds.get(1)), ys)
            upperRight = findCorner(x(bounds.get(2)), xs)
            upperLeft = findCorner(y(bounds.get(3)), ys)
            return {
                'upperLeft': upperLeft,
                'upperRight': upperRight,
                'lowerRight': lowerRight,
                'lowerLeft': lowerLeft
            }

        corners = findCorners(image)

        def viewAngles(img):
            maxDistanceToSceneEdge = 1000000
            maxSatelliteZenith = 7.5
            upperCenter = pointBetween(corners['upperLeft'], corners['upperRight'])
            lowerCenter = pointBetween(corners['lowerLeft'], corners['lowerRight'])
            slope = slopeBetween(lowerCenter, upperCenter)
            slopePerp = ee.Number(-1).divide(slope)
            img = set_name(img, 'viewAz',
                             ee.Image(ee.Number(math.pi / 2).subtract((slopePerp).atan())))

            leftLine = toLine(corners['upperLeft'], corners['lowerLeft'])
            rightLine = toLine(corners['upperRight'], corners['lowerRight'])
            leftDistance = ee.FeatureCollection(leftLine) \
                .distance(maxDistanceToSceneEdge)
            rightDistance = ee.FeatureCollection(rightLine) \
                .distance(maxDistanceToSceneEdge)
            viewZenith = rightDistance.multiply(maxSatelliteZenith * 2) \
                .divide(rightDistance.add(leftDistance)) \
                .subtract(maxSatelliteZenith)
            img = set_name(img, 'viewZen',
                             viewZenith.multiply(math.pi).divide(180))

            return img

        image = viewAngles(image)

        def solarPosition(img):
            # Ported from http://pythonfmask.org/en/latest/_modules/fmask/landsatangles.html
            date = ee.Date(ee.Number(img.get('system:time_start')))
            secondsInHour = 3600
            img = set_name(img, 'longDeg',
                           ee.Image.pixelLonLat().select('longitude'))

            img = set_name(img, 'latRad',
                           ee.Image.pixelLonLat().select('latitude') \
                           .multiply(math.pi).divide(180))

            img = set_name(img, 'hourGMT',
                           ee.Number(date.getRelative('second', 'day')).divide(secondsInHour))

            img = set_name(img, 'jdp', # Julian Date Proportion
                           date.getFraction('year'))

            img = set_name(img, 'jdpr', # Julian Date Proportion in Radians
                           'i.jdp * 2 * {pi}')


            img = set_name(img, 'meanSolarTime',
                           'i.hourGMT + i.longDeg / 15')

            img = set_name(img, 'localSolarDiff',
                           '(0.000075 + 0.001868 * cos(i.jdpr) - 0.032077 * sin(i.jdpr)' +
                           '- 0.014615 * cos(2 * i.jdpr) - 0.040849 * sin(2 * i.jdpr))' +
                           '* 12 * 60 / {pi}')

            img = set_name(img, 'trueSolarTime',
                           'i.meanSolarTime + i.localSolarDiff / 60 - 12')

            img = set_name(img, 'angleHour',
                           'i.trueSolarTime * 15 * {pi} / 180')

            img = set_name(img, 'delta',
                           '0.006918'+
                           '- 0.399912 * cos(1 * i.jdpr) + 0.070257 * sin(1 * i.jdpr)'+
                           '- 0.006758 * cos(2 * i.jdpr) + 0.000907 * sin(2 * i.jdpr)'+
                           '- 0.002697 * cos(3 * i.jdpr) + 0.001480 * sin(3 * i.jdpr)')

            img = set_name(img, 'cosSunZen',
                           'sin(i.latRad) * sin(i.delta) ' +
                           '+ cos(i.latRad) * cos(i.delta) * cos(i.angleHour)')

            img = set_name(img, 'sunZen',
                           'acos(i.cosSunZen)')

            img = set_name(img, 'sinSunAzSW',
                           toImage(img, 'cos(i.delta) * sin(i.angleHour) / sin(i.sunZen)') \
                           .clamp(-1, 1))

            img = set_name(img, 'cosSunAzSW',
                           '(-cos(i.latRad) * sin(i.delta)' +
                           '+ sin(i.latRad) * cos(i.delta) * cos(i.angleHour)) / sin(i.sunZen)')

            img = set_name(img, 'sunAzSW',
                           'asin(i.sinSunAzSW)')

            img = setIf(img, 'sunAzSW',
                        'i.cosSunAzSW <= 0',
                        '{pi} - i.sunAzSW',
                        'i.sunAzSW')

            img = setIf(img, 'sunAzSW',
                        'i.cosSunAzSW > 0 and i.sinSunAzSW <= 0',
                        '2 * {pi} + i.sunAzSW',
                        'i.sunAzSW')

            img = set_name(img, 'sunAz',
                           'i.sunAzSW + {pi}')

            img = setIf(img, 'sunAz',
                        'i.sunAz > 2 * {pi}',
                        'i.sunAz - 2 * {pi}',
                        'i.sunAz')

            return img

        image = solarPosition(image)

        def sunZenOut(img):
            # https://nex.nasa.gov/nex/static/media/publication/HLS.v1.0.UserGuide.pdf
            img = set_name(img, 'centerLat',
                             ee.Number(
                                 ee.Geometry(img.get('system:footprint')) \
                                     .bounds().centroid(30).coordinates().get(0)) \
                             .multiply(math.pi).divide(180))
            img = set_name(img, 'sunZenOut',
                             '(31.0076' +
                             '- 0.1272 * i.centerLat' +
                             '+ 0.01187 * pow(i.centerLat, 2)' +
                             '+ 2.40E-05 * pow(i.centerLat, 3)' +
                             '- 9.48E-07 * pow(i.centerLat, 4)' +
                             '- 1.95E-09 * pow(i.centerLat, 5)' +
                             '+ 6.15E-11 * pow(i.centerLat, 6)) * {pi}/180')

            return img

        image = sunZenOut(image)
        image = set_name(image, 'relativeSunViewAz', 'i.sunAz - i.viewAz')

        def cosPhaseAngle(img, name, sunZen, viewZen, relativeSunViewAz):
            args = {
                'sunZen': sunZen,
                'viewZen': viewZen,
                'relativeSunViewAz': relativeSunViewAz}

            img = set_name(img, name,
                             toImage(img, 'cos({sunZen}) * cos({viewZen})' +
                                     '+ sin({sunZen}) * sin({viewZen}) * cos({relativeSunViewAz})',
                                     args).clamp(-1, 1))
            return img

        def rossThick(img, bandName, sunZen, viewZen, relativeSunViewAz):
            args = {'sunZen': sunZen,
                    'viewZen': viewZen,
                    'relativeSunViewAz': relativeSunViewAz}

            img = cosPhaseAngle(img, 'cosPhaseAngle', sunZen, viewZen,
                                  relativeSunViewAz)
            img = set_name(img, 'phaseAngle',
                             'acos(i.cosPhaseAngle)')

            img = set_name(img, bandName,
                             '(({pi}/2 - i.phaseAngle) * i.cosPhaseAngle + sin(i.phaseAngle)) ' +
                             '/ (cos({sunZen}) + cos({viewZen})) - {pi}/4', args)

            return img

        image = rossThick(image, 'kvol', 'i.sunZen', 'i.viewZen', 'i.relativeSunViewAz')
        image = rossThick(image, 'kvol0', 'i.sunZenOut', 0, 0)

        def anglePrime(img, name, angle):
            args = {'b/r': 1, 'angle': angle}
            img = set_name(img, 'tanAnglePrime',
                             '{b/r} * tan({angle})', args)
            img = setIf(img, 'tanAnglePrime', 'i.tanAnglePrime < 0', 0)
            img = set_name(img, name,
                             'atan(i.tanAnglePrime)')
            return img

        def liThin(img, bandName, sunZen, viewZen, relativeSunViewAz):
            # From https://modis.gsfc.nasa.gov/data/atbd/atbd_mod09.pdf
            args = {
                'sunZen': sunZen,
                'viewZen': viewZen,
                'relativeSunViewAz': relativeSunViewAz,
                'h/b': 2,
            }

            img = anglePrime(img, 'sunZenPrime', sunZen)
            img = anglePrime(img, 'viewZenPrime', viewZen)
            img = cosPhaseAngle(img, 'cosPhaseAnglePrime', 'i.sunZenPrime',
                                  'i.viewZenPrime', relativeSunViewAz)
            img = set_name(img, 'distance',
                             'sqrt(pow(tan(i.sunZenPrime), 2) + pow(tan(i.viewZenPrime), 2)' +
                             '- 2 * tan(i.sunZenPrime) * tan(i.viewZenPrime)'+
                             '* cos({relativeSunViewAz}))',
                             args)
            img = set_name(img, 'temp',
                             '1/cos(i.sunZenPrime) + 1/cos(i.viewZenPrime)')
            img = set_name(img, 'cosT',
                             toImage(img, '{h/b} * sqrt(pow(i.distance, 2)'+
                                     '+ pow(tan(i.sunZenPrime) * tan(i.viewZenPrime)'+
                                     '* sin({relativeSunViewAz}), 2))/i.temp', args) \
                             .clamp(-1, 1))
            img = set_name(img, 't', 'acos(i.cosT)')
            img = set_name(img, 'overlap',
                             '(1/{pi}) * (i.t - sin(i.t) * i.cosT) * (i.temp)')
            img = setIf(img, 'overlap', 'i.overlap > 0', 0)
            img = set_name(img, bandName,
                             'i.overlap - i.temp' +
                             '+ (1/2) * (1 + i.cosPhaseAnglePrime) * (1/cos(i.sunZenPrime))'+
                             '* (1/cos(i.viewZenPrime))')

            return img

        image = liThin(image, 'kgeo', 'i.sunZen', 'i.viewZen', 'i.relativeSunViewAz')
        image = liThin(image, 'kgeo0', 'i.sunZenOut', 0, 0)

        def brdf(img, bandName, kvolBand, kgeoBand, coefficients):
            args = merge(coefficients, {
                'kvol': '3 * i.' + kvolBand, # check this multiplication factor. Is there an 'optimal' value?  Without a factor here, there is not enough correction.
                'kgeo': 'i.' + kgeoBand
            })

            img = set_name(img, bandName,
                           '{fiso} + {fvol} * {kvol} + {fgeo} * {kvol}', args)
            return img

        def applyCFactor(img, bandName, coefficients):
            img = brdf(img, 'brdf', 'kvol', 'kgeo', coefficients)
            img = brdf(img, 'brdf0', 'kvol0', 'kgeo0', coefficients)
            img = set_name(img, 'cFactor',
                           'i.brdf0 / i.brdf', coefficients)
            img = set_name(img, bandName,
                           '{bandName} * i.cFactor', {'bandName': 'i.' + bandName})
            return img

        def adjustBands(img):
            """ apply cFactor per band """
            for bandName in coefficientsByBand:
                coefficients = coefficientsByBand[bandName]
                img = applyCFactor(img, bandName, coefficients)
            return img

        image = adjustBands(image).select(inputBandNames)

        return original.addBands(image, overwrite=True)



# coding=utf-8

import ee
from geetools import cloud_mask
from geetools.tools.image import getValue
from . import TEST_CLOUD_IMAGES

# Initialize
ee.Initialize()

p_cloud = ee.Geometry.Point([-72.35439658020805, -42.909877084944945])
p_shadow = ee.Geometry.Point([-72.40366193088799, -42.8993149292846])
p_snow = ee.Geometry.Point([-72.31921773694792, -42.92788613999289])
p_cirrus = ee.Geometry.Point([-72.36041235428117, -42.90549801791416])

image = ee.Image(TEST_CLOUD_IMAGES['S2'])


def test_clouds():
    masked = cloud_mask.applyHollstein(image, ['cloud'])
    vals = getValue(masked, p_cloud, 30)

    assert vals.get("B1").getInfo() == None


def test_cirrus():
    masked = cloud_mask.applyHollstein(image, ['cirrus'])
    vals = getValue(masked, p_cirrus, 30)

    assert vals.get("B1").getInfo() == None


def test_shadow():
    masked = cloud_mask.applyHollstein(image, ['shadow'])
    vals = getValue(masked, p_shadow, 30)

    assert vals.get("B1").getInfo() == None


def test_snow():
    masked = cloud_mask.applyHollstein(image, ['snow'])
    vals = getValue(masked, p_snow, 30)

    assert vals.get("B1").getInfo() == None


# coding=utf-8
import ee
from . import utils
import json
import csv
from .. import tools


def fromShapefile(filename, crs=None, start=None, end=None):
    """ Convert an ESRI file (.shp and .dbf must be present) to a
    ee.FeatureCollection

    At the moment only works for shapes with less than 1000 records and doesn't
    handle complex shapes.

    :param filename: the name of the filename. If the shape is not in the
        same path than the script, specify a path instead.
    :type filename: str
    :param start:
    :return: the FeatureCollection
    :rtype: ee.FeatureCollection
    """
    import shapefile

    wgs84 = ee.Projection('EPSG:4326')
    # read the filename
    reader = shapefile.Reader(filename)
    fields = reader.fields[1:]
    field_names = [field[0] for field in fields]
    field_types = [field[1] for field in fields]
    types = dict(zip(field_names, field_types))
    features = []

    projection = utils.getProjection(filename) if not crs else crs
    # catch a string with format "EPSG:XXX"
    if isinstance(projection, str):
        if 'EPSG:' in projection:
            projection = projection.split(':')[1]
    projection = 'EPSG:{}'.format(projection)

    # filter records with start and end
    start = start if start else 0
    if not end:
        records = reader.shapeRecords()
        end = len(records)
    else:
        end = end + 1

    if (end-start)>1000:
        msg = "Can't process more than 1000 records at a time. Found {}"
        raise ValueError(msg.format(end-start))

    for i in range(start, end):
        # atr = dict(zip(field_names, sr.record))
        sr = reader.shapeRecord(i)
        atr = {}
        for fld, rec in zip(field_names, sr.record):
            fld_type = types[fld]
            if fld_type == 'D':
                value = ee.Date(rec.isoformat()).millis().getInfo()
            elif fld_type in ['C', 'N', 'F']:
                value = rec
            else:
                continue
            atr[fld] = value
        geom = sr.shape.__geo_interface__
        if projection is not None:
            geometry = ee.Geometry(geom, projection) \
                .transform(wgs84, 1)
        else:
            geometry = ee.Geometry(geom)
        feat = ee.Feature(geometry, atr)
        features.append(feat)

    return ee.FeatureCollection(features)


def fromGeoJSON(filename=None, data=None, crs=None):
    """ Create a list of Features from a GeoJSON file. Return a python tuple
    with ee.Feature inside. This is due to failing when attempting to create a
    FeatureCollection (Broken Pipe ERROR) out of the list. You can try creating
    it yourself casting the result of this function to a ee.List or using it
    directly as a FeatureCollection argument.

    :param filename: the name of the file to load
    :type filename: str
    :param crs: a coordinate reference system in EPSG format. If not specified
        it will try to get it from the geoJSON, and if not there it will rise
        an error
    :type: crs: str
    :return: a tuple of features.
    """
    if filename:
        with open(filename, 'r') as geoj:
            content = geoj.read()
            geodict = json.loads(content)
    else:
        geodict = data

    features = []
    # Get crs from GeoJSON
    if not crs:
        filecrs = geodict.get('crs')
        if filecrs:
            name = filecrs.get('properties').get('name')
            splitcrs = name.split(':')
            cleancrs = [part for part in splitcrs if part]
            try:
                if cleancrs[-1] == 'CRS84':
                    crs = 'EPSG:4326'
                elif cleancrs[-2] == 'EPSG':
                    crs = '{}:{}'.format(cleancrs[-2], cleancrs[-1])
                else:
                    raise ValueError('{} not recognized'.format(name))
            except IndexError:
                raise ValueError('{} not recognized'.format(name))
        else:
            crs = 'EPSG:4326'

    for n, feat in enumerate(geodict.get('features')):
        properties = feat.get('properties')
        geom = feat.get('geometry')
        ty = geom.get('type')
        coords = geom.get('coordinates')
        if ty == 'GeometryCollection':
            ee_geom = utils.GEOMETRY_TYPES.get(ty)(geom, opt_proj=crs)
        else:
            if ty == 'Polygon':
                coords = utils.removeZ(coords) if utils.hasZ(coords) else coords
            ee_geom = utils.GEOMETRY_TYPES.get(ty)(coords, proj=ee.Projection(crs))
        ee_feat = ee.feature.Feature(ee_geom, properties)
        features.append(ee_feat)

    return tuple(features)


def fromKML(filename=None, data=None, crs=None, encoding=None):
    """ Create a list of Features from a KML file. Return a python tuple
    with ee.Feature inside. This is due to failing when attempting to create a
    FeatureCollection (Broken Pipe ERROR) out of the list. You can try creating
    it yourself casting the result of this function to a ee.List or using it
    directly as a FeatureCollection argument.

    :param filename: the name of the file to load
    :type filename: str
    :param crs: a coordinate reference system in EPSG format. If not specified
        it will try to get it from the geoJSON, and if not there it will rise
        an error
    :type: crs: str
    :return: a tuple of features.
    """
    geojsondict = utils.kmlToGeoJsonDict(filename, data, encoding)
    features = geojsondict['features']

    for feat in features:
        # remove styleUrl
        prop = feat['properties']
        if 'styleUrl' in prop:
            prop.pop('styleUrl')

        # remove Z value if needed
        geom = feat['geometry']
        ty = geom['type']
        if ty == 'GeometryCollection':
            geometries = geom['geometries']
            for g in geometries:
                c = g['coordinates']
                utils.removeZ(c)
        else:
            coords = geom['coordinates']
            utils.removeZ(coords)

    return fromGeoJSON(data=geojsondict, crs=crs)


def toDict(collection, split_at=4000):
    """ Get the FeatureCollection as a dict object """
    size = collection.size()
    condition = size.gte(4999)

    def greater():
        size = collection.size()
        seq = tools.ee_list.sequence(0, size, split_at)
        limits = ee.List.zip(seq.slice(1), seq)

        def over_limits(n):
            n = ee.List(n)
            ini = ee.Number(n.get(0))
            end = ee.Number(n.get(1))
            return ee.FeatureCollection(collection.toList(ini, end))

        return limits.map(over_limits)

    collections = ee.List(
        ee.Algorithms.If(condition,
                         greater(),
                         ee.List([collection])))

    collections_size = collections.size().getInfo()

    col = ee.FeatureCollection(collections.get(0))
    content = col.getInfo()
    feats = content['features']

    for i in range(0, collections_size):
        c = ee.FeatureCollection(collections.get(i))
        content_c = c.getInfo()
        feats_c = content_c['features']
        feats = feats + feats_c

    content['features'] = feats

    return content


def toGeoJSON(collection, name, path=None, split_at=4000):
    """ Export a FeatureCollection to a GeoJSON file

    :param collection: The collection to export
    :type collection: ee.FeatureCollection
    :param name: name of the resulting file
    :type name: str
    :param path: The path where to save the file. If None, will be saved
        in the current folder
    :type path: str
    :param split_at: limit to avoid an EE Exception
    :type split_at: int
    :return: A GeoJSON (.geojson) file.
    :rtype: file
    """
    import json
    import os

    if not path:
        path = os.getcwd()

    # name
    if name[-8:-1] != '.geojson':
        fname = name+'.geojson'

    content = toDict(collection, split_at)

    with open(os.path.join(path, fname), 'w') as thefile:
        thefile.write(json.dumps(content))

    return thefile


def toCSV(collection, filename, split_at=4000):
    """ Alternative to download a FeatureCollection as a CSV """
    d = toDict(collection, split_at)

    fields = list(d['columns'].keys())
    fields.append('geometry')

    features = d['features']

    ext = filename[-4:]
    if ext != '.csv':
        filename += '.csv'

    with open(filename, 'w') as thecsv:
        writer = csv.DictWriter(thecsv, fields)

        writer.writeheader()
        # write rows
        for feature in features:
            properties = feature['properties']
            fid = feature['id']
            geom = feature['geometry']['type']

            # match fields
            properties['system:index'] = fid
            properties['geometry'] = geom

            # write row
            writer.writerow(properties)

        return thecsv


def toLocal(collection, filename, filetype=None, selectors=None, path=None):
    """ Download a FeatureCollection to a local file a CSV or geoJSON file.
    This uses a different method than `toGeoJSON` and `toCSV`

    :param filetype: The filetype of download, either CSV or JSON.
        Defaults to CSV.
    :param selectors: The selectors that should be used to determine which
        attributes will be downloaded.
    :param filename: The name of the file to be downloaded
    """
    if not filetype:
        filetype = 'CSV'

    url = collection.getDownloadURL(filetype, selectors, filename)
    thefile = utils.downloadFile(url, filename, filetype, path)
    return thefile


def toAsset(table, assetPath, name=None, create=True, verbose=False, **kwargs):
    """ This function can create folders and ImageCollections on the fly.
    The rest is the same to Export.image.toAsset. You can pass the same
    params as the original function

    :param table: the feature collection to upload
    :type table: ee.FeatureCollection
    :param assetPath: path to upload the image (only PATH, without
        filename)
    :type assetPath: str
    :param name: filename for the image (AssetID will be assetPath + name)
    :type name: str
    :return: the tasks
    :rtype: ee.batch.Task
    """

    # Check if the user is specified in the asset path
    is_user = (assetPath.split('/')[0] == 'users')
    if not is_user:
        user = ee.batch.data.getAssetRoots()[0]['id']
        assetPath = "{}/{}".format(user, assetPath)

    if create:
        # Recrusive create path
        path2create = assetPath #  '/'.join(assetPath.split('/')[:-1])
        utils.createAssets([path2create], 'Folder', True)

    # Asset ID (Path + name)
    assetId = '/'.join([assetPath, name])
    # Description
    description = utils.matchDescription(name)
    # Init task
    task = ee.batch.Export.table.toAsset(table, assetId=assetId,
                                         description=description, **kwargs)
    task.start()
    if verbose:
        print('Exporting {} to {}'.format(name, assetPath))

    return task


# coding=utf-8
""" Module holding tools for ee.Geometry """
import ee
import ee.data


UNBOUNDED = [[[-180.0, -90.0], [180.0, -90.0],
              [180.0, 90.0], [-180.0, 90.0],
              [-180.0, -90.0]]]

UNBOUNDED_2 = [[[None, None], [None, None],
                [None, None], [None, None],
                [None, None]]]


def isUnbounded(geometry):
    """ Check if the geometry is unbounded using the module's definition of
    unboundness. It's a client-side function. For server-side function use

    `ee.Geometry.isUnbounded()`
    """
    bounds = geometry.getInfo()['coordinates']
    check1 = bounds == UNBOUNDED
    check2 = bounds == UNBOUNDED_2

    if (check1 or check2):
        return True
    else:
        return False

UNBOUNDED = [[[-180.0, -90.0], [180.0, -90.0],
              [180.0, 90.0], [-180.0, 90.0],
              [-180.0, -90.0]]]

UNBOUNDED_2 = [[[None, None], [None, None],
                [None, None], [None, None],
                [None, None]]]


def isUnbounded(geometry):
    """ Check if the geometry is unbounded using the module's definition of
    unboundness. It's a client-side function. For server-side function use

    `ee.Geometry.isUnbounded()`
    """
    bounds = geometry.getInfo()['coordinates']
    check1 = bounds == UNBOUNDED
    check2 = bounds == UNBOUNDED_2

    if (check1 or check2):
        return True
    else:
        return False


def unpack(iterable):
    """ Helper function to unpack an iterable """
    unpacked = []
    for tt in iterable:
        for t in tt:
            unpacked.append(t)
    return unpacked


def getRegion(eeobject, bounds=False, error=1):
    """ Gets the region of a given geometry to use in exporting tasks. The
    argument can be a Geometry, Feature or Image

    :param eeobject: geometry to get region of
    :type eeobject: ee.Feature, ee.Geometry, ee.Image
    :param error: error parameter of ee.Element.geometry
    :return: region coordinates ready to use in a client-side EE function
    :rtype: json
    """
    def dispatch(geometry):
        info = geometry.getInfo()
        geomtype = info['type']
        if geomtype == 'GeometryCollection':
            geometries = info['geometries']
            region = []
            for geom in geometries:
                this_type = geom['type']
                if this_type in ['Polygon', 'Rectangle']:
                    region.append(geom['coordinates'][0])
                elif this_type in ['MultiPolygon']:
                    geometries2 = geom['coordinates']
                    region.append(unpack(geometries2))

        elif geomtype == 'MultiPolygon':
            subregion = info['coordinates']
            region = unpack(subregion)
        else:
            region = info['coordinates']

        return region

    # Geometry
    if isinstance(eeobject, ee.Geometry):
        geometry = eeobject.bounds() if bounds else eeobject
        region = dispatch(geometry)
    # Feature and Image
    elif isinstance(eeobject, (ee.Feature, ee.Image)):
        geometry = eeobject.geometry(error).bounds() if bounds else eeobject.geometry(error)
        region = dispatch(geometry)
    # FeatureCollection and ImageCollection
    elif isinstance(eeobject, (ee.FeatureCollection, ee.ImageCollection)):
        if bounds:
            geometry = eeobject.geometry(error).bounds()
        else:
            geometry = eeobject.geometry(error).dissolve()
        region = dispatch(geometry)
    # List
    elif isinstance(eeobject, list):
        condition = all([type(item) == list for item in eeobject])
        if condition:
            region = eeobject
    else:
        region = eeobject

    return region



# coding=utf-8
""" Functions for calculation indices """
FORMULAS = {
    'NDVI': '(NIR-RED)/(NIR+RED)',
    'EVI': 'G*((NIR-RED)/(NIR+(C1*RED)-(C2*BLUE)+L))',
    'NBR': '(NIR-SWIR2)/(NIR+SWIR2)',
    'NBR2': '(SWIR-SWIR2)/(SWIR+SWIR2)',
    'NDWI': '(NIR-SWIR)/(NIR+SWIR)'
}

AVAILABLE = FORMULAS.keys()

def compute(image, index, band_params, extra_params=None, bandname=None):
    if index not in AVAILABLE:
        raise ValueError('Index not available')

    if not bandname:
        bandname = index

    extra_params = extra_params if extra_params else {}

    formula = FORMULAS[index]

    band_params_mapped = {
        key: image.select([val]) for key, val in band_params.items()}
    band_params_mapped.update(extra_params)
    nd = image.expression(formula, band_params_mapped).rename(bandname)

    return nd

def ndvi(image, nir, red, bandname='ndvi'):
    """ Calculates NDVI index

    :USE:

    .. code:: python

        # use in a collection
        col = ee.ImageCollection(ID)
        ndvi = col.map(indices.ndvi("B4", "B3"))

        # use in a single image
        img = ee.Image(ID)
        ndvi = Indices.ndvi("NIR", "RED")(img)

    :param nir: name of the Near Infrared () band
    :type nir: str
    :param red: name of the red () band
    :type red: str
    :param addBand: if True adds the index band to the others, otherwise
        returns just the index band
    :type addBand: bool
    :return: The function to apply a map() over a collection
    :rtype: function
    """
    return compute(image, 'NDVI', {'NIR':nir, 'RED':red}, bandname=bandname)


def evi(image, nir, red, blue, G=2.5, C1=6, C2=7.5, L=1, bandname='evi'):
    """ Calculates EVI index

    :param nir: name of the Near Infrared () band
    :type nir: str
    :param red: name of the red () band
    :type red: str
    :param blue: name of the blue () band
    :type blue: str
    :param G: G coefficient for the EVI index
    :type G: float
    :param C1: C1 coefficient for the EVI index
    :type C1: float
    :param C2: C2 coefficient for the EVI index
    :type C2: float
    :param L: L coefficient for the EVI index
    :type L: float
    :param addBand: if True adds the index band to the others, otherwise
        returns just the index band
    :return: The function to apply a map() over a collection
    :rtype: function
    """
    G = float(G)
    C1 = float(C1)
    C2 = float(C2)
    L = float(L)

    return compute(image, 'EVI', {'NIR':nir, 'RED':red, 'BLUE':blue},
                   {'G':G, 'C1':C1, 'C2':C2, 'L':L}, bandname=bandname)


def nbr(image, nir, swir2, bandname='nbr'):
    """ Calculates NBR index

    :USE:

    .. code:: python

        # use in a collection
        col = ee.ImageCollection(ID)
        ndvi = col.map(indices.ndvi("B4", "B3"))

        # use in a single image
        img = ee.Image(ID)
        ndvi = Indices.ndvi("NIR", "RED")(img)

    :param nir: name of the Near Infrared () band
    :type nir: str
    :param red: name of the red () band
    :type red: str
    :param addBand: if True adds the index band to the others, otherwise
        returns just the index band
    :type addBand: bool
    :return: The function to apply a map() over a collection
    :rtype: function
    """
    return compute(image, 'NBR', {'NIR':nir, 'SWIR2':swir2}, bandname=bandname)


def nbr2(image, swir, swir2, bandname='nbr2'):
    """ Calculates NBR index

    :USE:

    .. code:: python

        # use in a collection
        col = ee.ImageCollection(ID)
        ndvi = col.map(indices.ndvi("B4", "B3"))

        # use in a single image
        img = ee.Image(ID)
        ndvi = Indices.ndvi("NIR", "RED")(img)

    :param nir: name of the Near Infrared () band
    :type nir: str
    :param red: name of the red () band
    :type red: str
    :param addBand: if True adds the index band to the others, otherwise
        returns just the index band
    :type addBand: bool
    :return: The function to apply a map() over a collection
    :rtype: function
    """
    return compute(image, 'NBR2', {'SWIR':swir, 'SWIR2':swir2}, bandname=bandname)


REL = {"NDVI": ndvi,
       "EVI": evi,
       "NBR": nbr,
       "NBR2": nbr2
       }


# coding=utf-8

import ee
from geetools import tools
ee.Initialize()


l8SR = ee.Image("LANDSAT/LC8_SR/LC82310772014043")

p_l8SR_cloud = ee.Geometry.Point([-65.8109, -25.0185])
p_l8SR_no_cloud = ee.Geometry.Point([-66.0306, -24.9338])


def test_addConstantBands():
    newimg = tools.image.addConstantBands(l8SR, 0, "a", "b", "c")
    vals = tools.image.getValue(newimg, p_l8SR_no_cloud, 30, 'client')

    assert vals["B1"] == 517
    assert vals["a"] == 0
    assert vals["b"] == 0
    assert vals["c"] == 0

    newimg2 = tools.image.addConstantBands(l8SR, a=0, b=1, c=2)
    vals2 = tools.image.getValue(newimg2, p_l8SR_no_cloud, 30, 'client')

    assert vals2["B1"] == 517
    assert vals2["a"] == 0
    assert vals2["b"] == 1
    assert vals2["c"] == 2

    newimg3 = tools.image.addConstantBands(l8SR, 0, "a", "b", "c", d=1, e=2)
    vals3 = tools.image.getValue(newimg3, p_l8SR_no_cloud, 30, 'client')

    assert vals3["B1"] == 517
    assert vals3["a"] == 0
    assert vals3["b"] == 0
    assert vals3["c"] == 0
    assert vals3["d"] == 1
    assert vals3["e"] == 2


def test_replace():
    # constant image with value 10 and name 'anyname'
    testband = ee.Image.constant(10).select([0],["anyname"])

    # replace band B1 of l8SR for the testband
    newimg = tools.image.replace(l8SR, "B1", testband)

    # get a value from a point with no clouds
    vals = tools.image.getValue(newimg, p_l8SR_no_cloud, 30, 'client')

    # assert
    assert vals["anyname"] == 10


def test_sumBands():
    newimg = tools.image.sumBands(l8SR, "added_bands", ("B1", "B2", "B3"))

    vals = tools.image.getValue(newimg, p_l8SR_no_cloud, 30, 'client')
    suma = int(vals["B1"]) + int(vals["B2"]) + int(vals["B3"])

    assert vals["added_bands"] == suma
    

def test_rename_bands():
    # Check original
    original_bands = l8SR.bandNames().getInfo()
    assert original_bands == ["B1", "B2", "B3", "B4", "B5", "B6", "B7",
                              "cfmask", "cfmask_conf"]

    # rename
    i = tools.image.renameDict(l8SR, {"B1": "BLUE", "B2": "GREEN"})

    # get value from point in a cloud free zone
    vals = tools.image.getValue(i, p_l8SR_no_cloud, 30, 'client')

    # get new band names
    bands = i.bandNames().getInfo()

    expected = ["BLUE", "GREEN", "B3", "B4", "B5", "B6", "B7",
                "cfmask", "cfmask_conf"]

    assert bands == expected
    assert vals["BLUE"] == 517


def test_parametrize():
    # parametrize
    newimg = tools.image.parametrize(l8SR, (0, 10000), (0, 1), ["B1", "B2"])

    # get a value
    vals = tools.image.getValue(newimg, p_l8SR_no_cloud, 30, 'client')

    # assert
    assert vals["B1"] == 517.0/10000
    assert vals["B2"] == 580.0/10000
    assert vals["B3"] == 824


def test_minscale():
    minscale = tools.image.minscale(l8SR).getInfo()

    # assert
    assert minscale == 30


def test_pass_prop():
    empty1 = ee.Image.constant(0)
    empty2 = ee.Image.constant(0).set("satellite", "None")

    pass1 = tools.image.passProperty(l8SR, empty1, ["satellite", "WRS_PATH"])
    pass2 = tools.image.passProperty(l8SR, empty2, ["satellite", "WRS_PATH"])

    assert pass1.get("satellite").getInfo() == "LANDSAT_8"
    assert pass2.get("satellite").getInfo() == "LANDSAT_8"
    assert pass1.get("WRS_PATH").getInfo() == 231
    assert pass2.get("WRS_PATH").getInfo() == 231



# coding=utf-8
""" ee.Array helpers """
import ee


def constant2DArray(width, height, value):
    """ Create an array of width x height with a fixed value """
    cols = ee.List.repeat(value, height)
    rows = ee.List.repeat(value, width)
    return ee.Array(cols.map(lambda n: rows))


def set2DValue(array, position, value):
    """ set the x and y values in a 2-D array """
    xpos = ee.Number(ee.List(position).get(0))
    ypos = ee.Number(ee.List(position).get(1))

    value = ee.Number(value)

    alist = ee.Array(array).toList()
    row = ee.List(alist.get(ypos))
    newrow = row.set(xpos, value)

    newlist = alist.set(ypos, newrow)

    return ee.Array(newlist)



# coding=utf-8

import ee
from geetools import cloud_mask
from geetools.tools.image import getValue
from . import TEST_CLOUD_IMAGES

# Initialize
ee.Initialize()

p_cloud = ee.Geometry.Point([-71.09680552342795, -42.65765342130262])
p_snow = ee.Geometry.Point([-71.34485556784499, -42.572002133174415])
p_shadow = ee.Geometry.Point([-71.48614073981918, -42.648269348224574])

image = ee.Image(TEST_CLOUD_IMAGES['L7TOA'])


def test_clouds():
    masked = cloud_mask.landsat457ToaBQA(['cloud'])(image)
    vals = getValue(masked, p_cloud, 30)

    assert vals.get("B1").getInfo() == None


def test_shadows():
    masked = cloud_mask.landsat457ToaBQA(['shadow'])(image)
    vals = getValue(masked, p_shadow, 30)

    assert vals.get("B1").getInfo() == None


def test_snow():
    masked = cloud_mask.landsat457ToaBQA(['snow'])(image)
    vals = getValue(masked, p_snow, 30)

    assert vals.get("B1").getInfo() == None


# coding=utf-8
""" Module holding tools for ee.Collection """
import ee


def enumerate(collection):
    """ Create a list of lists in which each element of the list is:
    [index, element]. For example, if you parse a FeatureCollection with 3
    Features you'll get: [[0, feat0], [1, feat1], [2, feat2]]

    :param collection: can be an ImageCollection or a FeatureCollection
    :return: ee.Collection
    """
    collist = collection.toList(collection.size())

    # first element
    ini = ee.Number(0)
    first_image = ee.Image(collist.get(0))
    first = ee.List([ini, first_image])

    start = ee.List([first])
    rest = collist.slice(1)

    def over_list(im, s):
        im = ee.Image(im)
        s = ee.List(s)
        last = ee.List(s.get(-1))
        last_index = ee.Number(last.get(0))
        index = last_index.add(1)
        return s.add(ee.List([index, im]))

    list = ee.List(rest.iterate(over_list, start))

    return list


def joinByProperty(primary, secondary, propertyField, outer=False):
    """ Join 2 collections by a given property field.
    It assumes ids are unique so uses ee.Join.saveFirst.
    It drops non matching features.

    Example:

    fc1 = ee.FeatureCollection([ee.Feature(geom=ee.Geometry.Point([0,0]),
                                       opt_properties={'id': 1, 'prop_from_fc1': 'I am from fc1'})])
    fc2 = ee.FeatureCollection([ee.Feature(geom=ee.Geometry.Point([0,0]),
                                       opt_properties={'id': 1, 'prop_from_fc2': 'I am from fc2'})])
    joined = joinById(fc1, fc2, 'id')
    print(joined.getInfo())

    """
    Filter = ee.Filter.equals(leftField=propertyField,
                              rightField=propertyField)
    join = ee.Join.saveFirst(matchKey='match', outer=outer)
    joined = join.apply(primary, secondary, Filter)
    def overJoined(feat):
        properties = feat.propertyNames()
        retain = properties.remove('match')
        match = ee.Feature(feat.get('match'))
        matchprop = match.toDictionary()
        return feat.select(retain).setMulti(matchprop)

    return joined.map(overJoined)



# coding=utf-8

import ee
from geetools.tools.geometry import getRegion
ee.Initialize()

l8SR = ee.Image("LANDSAT/LC8_SR/LC82310772014043")

p_l8SR_cloud = ee.Geometry.Point([-65.8109, -25.0185])
p_l8SR_no_cloud = ee.Geometry.Point([-66.0306, -24.9338])

list1 = ee.List([1, 2, 3, 4, 5])
list2 = ee.List([4, 5, 6, 7])


def test_getRegion():
    expected = [[0.0,0.0], [10.0, 0.0], [10.0, 10.0], [0.0, 10.0], [0.0, 0.0]]
    pol = ee.Geometry.Polygon(expected)
    feat = ee.Feature(pol)

    region_geom = getRegion(pol)
    region_feat = getRegion(feat)

    assert region_geom == [expected]
    assert region_feat == [expected]


# coding=utf-8
""" Dispatch methods for different EE Object types """
import ee


def belongToEE(eeobject):
    """ Determine if the parsed object belongs to the Earth Engine API """
    module = getattr(eeobject, '__module__', None)
    parent = module.split('.')[0] if module else None
    if parent == ee.__name__:
        return True
    else:
        return False


# GENERAL DISPATCHER
def dispatch(eeobject):
    """ General dispatcher """
    if belongToEE(eeobject):
        # DISPATCH!!
        if isinstance(eeobject, (ee.Image,)):
            return dispatchImage(eeobject)
        elif isinstance(eeobject, (ee.Date,)):
            return dispatchDate(eeobject)
        elif isinstance(eeobject, (ee.DateRange,)):
            return dispatchDaterange(eeobject)
        # ADD MORE ABOVE ME!
        else:
            info = eeobject.getInfo()
            return info
    else:
        info = str(eeobject)
        return info


def dispatchImage(image):
    """ Dispatch a Widget for an Image Object """
    info = image.getInfo()

    # IMAGE
    image_id = info['id'] if 'id' in info else 'No Image ID'
    prop = info.get('properties')
    bands = info.get('bands')
    bands_names = [band.get('id') for band in bands]

    # BAND PRECISION
    bands_precision = []
    for band in bands:
        data = band.get('data_type')
        if data:
            precision = data.get('precision')
            bands_precision.append(precision)

    # BAND CRS
    bands_crs = []
    for band in bands:
        crs = band.get('crs')
        bands_crs.append(crs)

    # BAND MIN AND MAX
    bands_min = []
    for band in bands:
        data = band.get('data_type')
        if data:
            bmin = data.get('min')
            bands_min.append(bmin)

    bands_max = []
    for band in bands:
        data = band.get('data_type')
        if data:
            bmax = data.get('max')
            bands_max.append(bmax)

    bands = {}
    for name, pres, crs, minval, maxval in zip(
            bands_names, bands_precision, bands_crs, bands_min, bands_max):
        bands[name] = dict(precision=pres, crs=crs, min=minval, max=maxval)

    return dict(id=image_id, bands=bands, properties=prop)


def dispatchDate(date):
    """ Dispatch a ee.Date """
    info = date.format().getInfo()

    return info


def dispatchDaterange(daterange):
    """ Dispatch a DateRange """
    start = daterange.start().format().getInfo()
    end = daterange.end().format().getInfo()
    value = '{} to {}'.format(start, end)

    return value



# coding=utf-8
import ee
import os
from . import utils
from .. import tools
from ..utils import makeName


def toLocal(image, name=None, path=None, scale=None, region=None,
            dimensions=None, toFolder=True):
    """ Download an Image to your hard drive

    :param image: the image to download
    :type image: ee.Image
    :param path: the path to download the image. If None, it will be
        downloaded to the same folder as the script is
    :type path: str
    :param name: name of the file
    :type name: str
    :param scale: scale of the image to download. If None, tries to get it.
    :type scale: int
    :param region: region to from where to download the image. If None,
        will be the image region
    :type region: ee.Geometry
    :param
    """
    # TODO: checkExist
    try:
        import zipfile
    except:
        raise ValueError(
            'zipfile module not found, install it using '
            '`pip install zipfile`')

    name = name if name else image.id().getInfo()

    scale = scale if scale else int(tools.image.minscale(image).getInfo())

    if region:
        region = tools.geometry.getRegion(region)
    else:
        region = tools.geometry.getRegion(image)

    params = {'region': region,
              'scale': scale}

    if dimensions:
        params = params.update({'dimensions': dimensions})

    url = image.getDownloadURL(params)

    ext = 'zip'

    utils.downloadFile(url, name, ext)

    filename = '{}.{}'.format(name, ext)

    original_filepath = os.path.join(os.getcwd(), filename)

    if path:
        filepath = os.path.join(path, filename)
        os.rename(original_filepath, filepath)
    else:
        path = os.getcwd()
        filepath = os.path.join(path, filename)

    try:
        zip_ref = zipfile.ZipFile(filepath, 'r')

        if toFolder:
            finalpath = os.path.join(path, name)
        else:
            finalpath = path

        zip_ref.extractall(finalpath)
        zip_ref.close()
    except:
        raise


def toAsset(image, assetPath, name=None, to='Folder', scale=None,
            region=None, create=True, verbose=False, **kwargs):
    """ This function can create folders and ImageCollections on the fly.
    The rest is the same to Export.image.toAsset. You can pass the same
    params as the original function

    :param image: the image to upload
    :type image: ee.Image
    :param assetPath: path to upload the image (only PATH, without
        filename)
    :type assetPath: str
    :param name: filename for the image (AssetID will be assetPath + name)
    :type name: str
    :param to: where to save the image. Options: 'Folder' or
        'ImageCollection'
    :param region: area to upload. Defualt to the footprint of the first
        image in the collection
    :type region: ee.Geometry.Rectangle or ee.Feature
    :param scale: scale of the image (side of one pixel)
        (Landsat resolution)
    :type scale: int
    :param dataType: as downloaded images **must** have the same data type
        in all bands, you have to set it here. Can be one of: "float",
        "double", "int", "Uint8", "Int8" or a casting function like
        *ee.Image.toFloat*
    :type dataType: str
    :param notebook: display a jupyter notebook widget to monitor the task
    :type notebook: bool
    :return: the tasks
    :rtype: ee.batch.Task
    """
    # Convert data type
    # image = utils.convertDataType(dataType)(image)

    # Check if the user is specified in the asset path
    is_user = (assetPath.split('/')[0] == 'users')
    if not is_user:
        user = ee.batch.data.getAssetRoots()[0]['id']
        assetPath = "{}/{}".format(user, assetPath)

    # description = kwargs.get('description', image.id().getInfo())
    # Set scale
    scale = scale if scale else int(tools.image.minscale(image).getInfo())

    if create:
        # Recrusive create path
        path2create = assetPath #  '/'.join(assetPath.split('/')[:-1])
        utils.createAssets([path2create], to, True)

    # Region
    region = tools.geometry.getRegion(region)
    # Name
    name = name if name else image.id().getInfo()
    # Asset ID (Path + name)
    assetId = '/'.join([assetPath, name])
    # Description
    description = utils.matchDescription(name)
    # Init task
    task = ee.batch.Export.image.toAsset(image, assetId=assetId,
                                         region=region, scale=scale,
                                         description=description,
                                         **kwargs)
    task.start()
    if verbose:
        print('Exporting {} to {}'.format(name, assetPath))

    return task


def toDriveByFeature(image, collection, folder, namePattern, datePattern=None,
                     scale=1000, dataType="float", verbose=False, **kwargs):
    """ Export an image clipped by features (Polygons). You can use the
    same arguments as the original function ee.batch.export.image.toDrive

    :Parameters:
    :param image: image to clip
    :type image: ee.Image
    :param collection: feature collection
    :type collection: ee.FeatureCollection
    :param folder: same as ee.Export
    :type folder: str
    :param namePattern: a name pattern using image and/or feature properties between
        brakets. Example: '{ID} {a_feat_prop} {an_image_prop}'
    :type namePattern: str
    :param datePattern: a date pattern to use for {system_date} pattern in name
    :type datePattern: str
    :param scale: same as ee.Export. Default to 1000
    :type scale: int
    :param dataType: as downloaded images **must** have the same data
    type in all
        bands, you have to set it here. Can be one of: "float",
        "double", "int",
        "Uint8", "Int8" or a casting function like *ee.Image.toFloat*
    :type dataType: str

    :return: a list of all tasks (for further processing/checking)
    :rtype: list
    """
    collist = collection.toList(collection.size())
    tasklist = []
    i = 0
    while True:
        try:
            feat = ee.Feature(collist.get(i))
            props = feat.toDictionary()

            n = makeName(image, namePattern, datePattern)
            n = tools.string.format(n, props)
            n = n.getInfo()
            n = n.replace('{','').replace('}','')
            desc = utils.matchDescription(n)

            # convert data type
            image = utils.convertDataType(dataType)(image)

            region = tools.geometry.getRegion(feat)

            task = ee.batch.Export.image.toDrive(
                image=image,
                description=desc,
                folder=folder,
                fileNamePrefix=n,
                region=region,
                scale=scale, **kwargs)

            task.start()
            if verbose:
                print("exporting '{}' to '{}' folder in GDrive".format(n, folder))
            tasklist.append(task)

            i += 1
        except Exception as e:
            error = str(e).split(':')
            if error[0] == 'List.get':
                break
            else:
                raise e


def qgisCode(image, visParams=None, name=None, namePattern=None,
             datePattern=None):
    QGIS_IMG_CODE = """name = '{name}'
url = '{url}'
urlWithParams = "type=xyz&url={{}}".format(url)
rlayer = QgsRasterLayer(urlWithParams, name, "wms")
if rlayer.isValid():
    QgsProject.instance().addMapLayer(rlayer)
else:
    print("invalid layer")
"""
    url = tools.image.getTileURL(image, visParams)
    if namePattern:
        name = makeName(image, namePattern, datePattern)
    else:
        name = name or 'no_name_img'
    return QGIS_IMG_CODE.format(name=name, url=url)


def toQGIS(image, visParams=None, name=None, filename=None, path=None,
           replace=True, verbose=False):
    """ Download a python file to import from QGIS """
    code = qgisCode(image, visParams, name)
    path = path or os.getcwd()
    # Check extension
    if filename:
        ext = filename.split('.')[-1]
        if ext != 'py':
            filename += '.py'
    else:
        filename = 'qgis2ee'
    # add _qgis_ to filename
    splitted = filename.split('.')[:-1]
    noext = '.'.join(splitted)
    filename = '{}_qgis_'.format(noext)
    # process
    finalpath = os.path.join(path, filename)
    finalpath = '{}.py'.format(finalpath)
    if not os.path.exists(finalpath) or replace:
        with open(finalpath, 'w+') as thefile:
            thefile.write(code)
        return thefile
    else:
        return None



""" EE Element. Common methods between Feature and Image """
import ee


def fillNull(Element, proxy=-999):
    """ Fill null values of an Element's properties with a proxy value """
    todict = Element.toDictionary()
    null = Element.propertyNames().removeAll(todict.keys()).remove('system:index')
    null_list = ee.List.repeat(proxy, null.size())
    null_dict = ee.Dictionary.fromLists(null, null_list)
    return Element.setMulti(null_dict)



""" EE Computed Object """
import ee

STRING = 'String'
INTEGER = 'Integer'
FLOAT = 'Float'
IMAGE = 'Image'
IMAGECOLLECTION = 'ImageCollection'
FEATURE = 'Feature'
GEOMETRY = 'Geometry'


def _isType(ComputedObject, checktype):
    """ Return 1 if the element is the passed type or 0 if not """
    otype = ee.Algorithms.ObjectType(ComputedObject)
    return otype.compareTo(checktype).eq(0)


def isString(ComputedObject):
    return _isType(ComputedObject, STRING)


def isInteger(ComputedObject):
    return _isType(ComputedObject, INTEGER)


def isFloat(ComputedObject):
    return _isType(ComputedObject, FLOAT)


def isNumber(ComputedObject):
    return isFloat(ComputedObject).Or(isInteger(ComputedObject))


def isImage(ComputedObject):
    return _isType(ComputedObject, IMAGE)


def isImageCollection(ComputedObject):
    return _isType(ComputedObject, IMAGECOLLECTION)


def isFeature(ComputedObject):
    return _isType(ComputedObject, FEATURE)


def isGeometry(ComputedObject):
    return _isType(ComputedObject, GEOMETRY)



# coding=utf-8
""" Bit Reader module """
import ee
import ee.data
from . import tools


class BitReader(object):
    """ Bit Reader

    Initializes with parameter `options`, which must be a dictionary with
    the following format:

    keys must be a str with the bits places, example: '0-1' means bit 0
    and bit 1

    values must be a dictionary with the bit value as the key and the category
    (str) as value. Categories must be unique.

    - Encode: given a category/categories return a list of possible values
    - Decode: given a value return a list of categories

    Example:
        MOD09 (http://modis-sr.ltdri.org/guide/MOD09_UserGuide_v1_3.pdf)
        (page 28, state1km, 16 bits):

        ```
        options = {
         '0-1': {0:'clear', 1:'cloud', 2:'mix'},
         '2-2': {1:'shadow'},
         '8-9': {1:'small_cirrus', 2:'average_cirrus', 3:'high_cirrus'}
         }

        reader = BitReader(options, 16)

        print(reader.decode(204))
        ```
        >>['shadow', 'clear']
        ```
        print(reader.match(204, 'cloud')
        ```
        >>False

    """
    @staticmethod
    def getBin(bit, nbits=None, shift=0):
        """ from https://stackoverflow.com/questions/699866/python-int-to-binary """
        pure = bin(bit)[2:]

        if not nbits:
            nbits = len(pure)

        lpure = len(pure)
        admited_shift = nbits-lpure
        if admited_shift < 0:
            mje = 'the number of bits must be more than the bits' \
                  ' representation of the number. {} ({}) cant be' \
                  ' represented in {} bits'
            raise ValueError(mje.format(pure, bit, nbits))

        if shift > admited_shift:
            mje = 'cant shift {} places for bit {} ({})'
            raise ValueError(mje.format(shift, pure, bit))

        if shift:
            shifted = bin(int(pure, 2)<<shift)[2:]
        else:
            shifted = pure
        return shifted.zfill(nbits)

    @staticmethod
    def decodeKey(key):
        """ decodes an option's key into a list """
        if isinstance(key, (str,)):
            bits = key.split('-')

            try:
                ini = int(bits[0])
                if len(bits) == 1:
                    end = ini
                else:
                    end = int(bits[1])
            except:
                mje = 'keys must be with the following format "bit-bit", ' \
                      'example "0-1" (found {})'
                raise ValueError(mje.format(key))

            bits_list = range(ini, end+1)
            return bits_list
        elif isinstance(key, (int, float)):
            value = int(key)
            return (value, value+1)

    def __init__(self, options, bit_length=None):
        self.options = options

        def allBits():
            ''' get a list of all bits and check consistance '''
            all_values  = [x for key in options.keys() for x in self.decodeKey(key)]
            for val in all_values:
                n = all_values.count(val)
                if n>1:
                    mje = "bits must not overlap. Example: {'0-1':.., " \
                          "'2-3':..} and NOT {'0-1':.., '1-3':..}"
                    raise ValueError(mje)
            return all_values

        ## Check if categories repeat and create property all_categories
        # TODO: reformat categories if find spaces or uppercases
        all_cat = []
        for key, val in self.options.items():
            for i, cat in val.items():
                if cat in all_cat:
                    msg = 'Classes must be unique, found "{}" twice'
                    raise ValueError(msg.format(cat))
                all_cat.append(cat)

        self.all_categories = all_cat
        ###

        all_values = allBits()

        self.bit_length = len(range(min(all_values), max(all_values)+1)) \
            if not bit_length else bit_length

        self.max = 2**self.bit_length

        info = {}
        for key, val in options.items():
            bits_list = self.decodeKey(key)
            bit_length_cat = len(bits_list)
            for i, cat in val.items():
                info[cat] = {'bit_length':bit_length_cat,
                             'lshift':bits_list[0],
                             'shifted': i
                             }
        self.info = info

    def encode(self, cat):
        """ Given a category, return the encoded value (only) """
        info = self.info[cat]
        lshift = info['lshift']
        decoded = info['shifted']

        shifted = decoded<<lshift
        return shifted

    def encodeBand(self, category, mask, name=None):
        """ Make an image in which all pixels have the value for the given
        category

        :param category: the category to encode
        :type category: str
        :param mask: the mask that indicates which pixels encode
        :type mask: ee.Image
        :param name: name of the resulting band. If None it'll be the same as
            'mask'
        :type name: str

        :return: A one band image
        :rtype: ee.Image
        """
        encoded = self.encode(category)

        if not name:
            name = mask.bandNames().get(0)

        image = tools.image.empty(encoded, [name])
        return image.updateMask(mask)


    def encodeAnd(self, *args):
        """ decodes a comination of the given categories. returns a list of
        possible values """
        first = args[0]
        values_first = self.encodeOne(first)

        def get_match(list1, list2):
            return [val for val in list2 if val in list1]

        result = values_first

        for cat in args[1:]:
            values = self.encodeOne(cat)
            result = get_match(result, values)
        return result

    def encodeOr(self, *args):
        """ decodes a comination of the given categories. returns a list of
        possible values """
        first = args[0]
        values_first = self.encodeOne(first)

        for cat in args[1:]:
            values = self.encodeOne(cat)
            for value in values:
                if value not in values_first:
                    values_first.append(value)

        return values_first

    def encodeNot(self, *args):
        """ Given a set of categories return a list of values that DO NOT
        match with any """
        result = []
        match = self.encodeOr(*args)
        for bit in range(self.max):
            if bit not in match:
                result.append(bit)
        return result

    def encodeOne(self, cat):
        """ Given a category, return a list of values that match it """
        info = self.info[cat]
        lshift = info['lshift']
        length = info['bit_length']
        decoded = info['shifted']

        result = []
        for bit in range(self.max):
            move = lshift+length
            rest = bit>>move<<move
            norest = bit-rest
            to_compare = norest>>lshift
            if to_compare == decoded:
                result.append(bit)
        return result

    def decode(self, value):
        """ given a value return a list with all categories """
        result = []
        for cat in self.all_categories:
            data = self.info[cat]
            lshift = data['lshift']
            length = data['bit_length']
            decoded = data['shifted']
            move = lshift+length
            rest = value>>move<<move
            norest = value-rest
            to_compare = norest>>lshift
            if to_compare == decoded:
                result.append(cat)
        return result

    def decodeImage(self, image, qa_band):
        """ Get an Image with one band per category in the Bit Reader

        :param bit_reader: the bit reader
        :type bit_reader: BitReader
        :param qa_band: name of the band that holds the bit information
        :type qa_band: str
        :return: the image with the decode bands added
        """
        options = ee.Dictionary(self.info)
        categories = ee.List(self.all_categories)

        def eachcat(cat, ini):
            ini = ee.Image(ini)
            qa = ini.select(qa_band)
            # get data for category
            data = ee.Dictionary(options.get(cat))
            lshift = ee.Number(data.get('lshift'))
            length = ee.Number(data.get('bit_length'))
            decoded = ee.Number(data.get('shifted'))
            # move = places to move bits right and left back
            move = lshift.add(length)
            # move bits right and left
            rest = qa.rightShift(move).leftShift(move)
            # subtract the rest
            norest = qa.subtract(rest)
            # right shift to compare with decoded data
            to_compare = norest.rightShift(lshift) ## Image
            # compare if is equal, return 0 if not equal, 1 if equal
            mask = to_compare.eq(decoded)
            # rename to the name of the category
            qa_mask = mask.select([0], [cat])

            return ini.addBands(qa_mask)
        return ee.Image(categories.iterate(eachcat, image)).select(categories)

    def match(self, value, category):
        """ given a value and a category return True if the value includes
        that category, else False """
        encoded = self.decode(value)
        return category in encoded


# coding=utf-8
"""
Tools for ee.Image
"""
from __future__ import absolute_import
import ee
import ee.data
import math
from . import ee_list, date, string
from ..utils import castImage
from ..ui import map as mapui


def _add_suffix_prefix(image, value, option, bands=None):
    """ Internal function to handle addPrefix and addSuffix """
    if bands:
        bands = ee.List(bands)

    addon = ee.String(value)

    allbands = image.bandNames()
    bands_ = ee.List(ee.Algorithms.If(bands, bands, allbands))

    def over_bands(band, first):
        all = ee.List(first)
        options = ee.Dictionary({
            'suffix': ee.String(band).cat(addon),
            'prefix': addon.cat(ee.String(band))
        })
        return all.replace(band, ee.String(options.get(option)))

    newbands = bands_.iterate(over_bands, allbands)
    newbands = ee.List(newbands)
    return image.select(allbands, newbands)


def addSuffix(image, suffix, bands=None):
    """ Add a suffix to the specified bands

    :param suffix: the value to add as a suffix
    :type suffix: str
    :param bands: the bands to apply the suffix. If None, suffix will fill
        all bands
    :type bands: list
    :rtype: ee.Image
    """
    return _add_suffix_prefix(image, suffix, 'suffix', bands)


def addPrefix(image, prefix, bands=None):
    """ Add a prefix to the specified bands

    :param prefix: the value to add as a prefix
    :type prefix: str
    :param bands: the bands to apply the prefix. If None, prefix will fill
        all bands
    :type bands: list
    :rtype: ee.Image
    """
    return _add_suffix_prefix(image, prefix, 'prefix', bands)


def bufferMask(image, radius=1.5, kernelType='square', units='pixels'):
    """ Make a buffer around the masked pixels """
    masked = image.mask().Not()
    buffer = masked.focal_max(radius, kernelType, units)
    return image.updateMask(buffer.Not())


def empty(value=0, names=None, from_dict=None):
    """ Create a constant image with the given band names and value, and/or
    from a dictionary of {name: value}

    :param names: list of names
    :type names: ee.List or list
    :param value: value for every band of the resulting image
    :type value: int or float
    :param from_dict: {name: value}
    :type from_dict: dict
    :rtype: ee.Image
    """
    image = ee.Image.constant(0)
    bandnames = ee.List([])
    if names:
        bandnames = names if isinstance(names, ee.List) else ee.List(names)
        def bn(name, img):
            img = ee.Image(img)
            newi = ee.Image(value).select([0], [name])
            return img.addBands(newi)
        image = ee.Image(bandnames.iterate(bn, image)) \
            .select(bandnames)

    if from_dict:
        bandnames = bandnames.cat(ee.List(from_dict.keys()))
        for name, value in from_dict.items():
            i = ee.Image(value).select([0], [name])
            image = image.addBands(i).select(bandnames)

    if not from_dict and not names:
        image = ee.Image.constant(value)

    return image


def emptyBackground(image, value=0):
    """ Make all background pixels (not only masked, but all over the world)
    take the parsed value """
    bnames = image.bandNames()
    emp = empty(value, bnames)
    return ee.Image(emp.where(image, image).copyProperties(
        source=image, properties=image.propertyNames()))


def emptyCopy(image, emptyValue=0, copyProperties=None, keepMask=False):
    """ Make an empty copy of the given image """
    footprint = image.geometry()
    emp = empty(emptyValue, image.bandNames())
    if copyProperties:
        emp = emp.copyProperties(source=image, properties=copyProperties)
    if keepMask:
        emp = emp.updateMask(image.mask())

    return ee.Image(ee.Algorithms.If(footprint.isUnbounded(),
                                     emp, emp.clip(footprint)))


def getValue(image, point, scale=None, side="server"):
    """ Return the value of all bands of the image in the specified point

    :param img: Image to get the info from
    :type img: ee.Image
    :param point: Point from where to get the info
    :type point: ee.Geometry.Point
    :param scale: The scale to use in the reducer. It defaults to 10 due to the
        minimum scale available in EE (Sentinel 10m)
    :type scale: int
    :param side: 'server' or 'client' side
    :type side: str
    :return: Values of all bands in the ponit
    :rtype: ee.Dictionary or dict
    """
    if scale:
        scale = int(scale)
    else:
        scale = 1

    type = point.getInfo()["type"]
    if type != "Point":
        raise ValueError("Point must be ee.Geometry.Point")

    result = image.reduceRegion(ee.Reducer.first(), point, scale)

    if side == 'server':
        return result
    elif side == 'client':
        return result.getInfo()
    else:
        raise ValueError("side parameter must be 'server' or 'client'")


def addMultiBands(imagesList):
    """ Image.addBands for many images. All bands from all images will be
    put together, so if there is one band with the same name in different
    images, the first occurrence will keep the name and the rest will have a
    number suffix ({band}_1, {band}_2, etc)

    :param imagesList: a list of images
    :type imagesList: list or ee.List
    :rtype: ee.Image
    """
    imagesList = ee.List(imagesList)
    first = ee.Image(imagesList.get(0))
    rest = imagesList.slice(1)

    def iteration(img, ini):
        ini = ee.Image(ini)
        img = ee.Image(img)
        return ini.addBands(img)

    return ee.Image(rest.iterate(iteration, first))


def renameDict(image, names):
    """ Renames bands of images using a dict

    :param names: matching names where key is original name and values the
        new name
    :type names: dict
    :rtype: ee.Image

    :EXAMPLE:

    .. code:: python

        image = ee.Image("LANDSAT/LC8_L1T_TOA_FMASK/LC82310902013344LGN00")
        p = ee.Geometry.Point(-71.72029495239258, -42.78997046797438)

        i = rename_bands({"B1":"BLUE", "B2":"GREEN"})

        print get_value(image, p)
        print get_value(i, p)

    >> {u'B1': 0.10094200074672699, u'B2': 0.07873955368995667, u'B3': 0.057160500437021255}
    >> {u'BLUE': 0.10094200074672699, u'GREEN': 0.07873955368995667, u'B3': 0.057160500437021255}
    """
    bandnames = image.bandNames()
    newnames = ee_list.replaceDict(bandnames, names)
    return image.select(bandnames, newnames)


def removeBands(image, bands):
    """ Remove the specified bands from an image """
    bnames = image.bandNames()
    bands = ee.List(bands)
    inter = ee_list.intersection(bnames, bands)
    diff = bnames.removeAll(inter)
    return image.select(diff)


def parametrize(image, range_from, range_to, bands=None, drop=False):
    """ Parametrize from a original **known** range to a fixed new range

    :param range_from: Original range. example: (0, 5000)
    :type range_from: tuple
    :param range_to: Fixed new range. example: (500, 1000)
    :type range_to: tuple
    :param bands: bands to parametrize. If *None* all bands will be
        parametrized.
    :type bands: list
    :param drop: drop the bands that will not be parametrized
    :type drop: bool

    :return: the parsed image with the parsed bands parametrized
    :rtype: ee.Image
    """
    original_range = range_from if isinstance(range_from, ee.List) \
        else ee.List(range_from)

    final_range = range_to if isinstance(range_to, ee.List) \
        else ee.List(range_to)

    # original min and max
    min0 = ee.Image.constant(original_range.get(0))
    max0 = ee.Image.constant(original_range.get(1))

    # range from min to max
    rango0 = max0.subtract(min0)

    # final min max images
    min1 = ee.Image.constant(final_range.get(0))
    max1 = ee.Image.constant(final_range.get(1))

    # final range
    rango1 = max1.subtract(min1)

    # all bands
    all = image.bandNames()

    # bands to parametrize
    if bands:
        bands_ee = ee.List(bands)
    else:
        bands_ee = image.bandNames()

    inter = ee_list.intersection(bands_ee, all)
    diff = ee_list.difference(all, inter)
    image_ = image.select(inter)

    # Percentage corresponding to the actual value
    percent = image_.subtract(min0).divide(rango0)

    # Taking count of the percentage of the original value in the original
    # range compute the final value corresponding to the final range.
    # Percentage * final_range + final_min

    final = percent.multiply(rango1).add(min1)

    if not drop:
        # Add the rest of the bands (no parametrized)
        final = image.select(diff).addBands(final)

    # return passProperty(image, final, 'system:time_start')
    return ee.Image(final.copyProperties(source=image))


def sumBands(image, name="sum", bands=None):
    """ Adds all *bands* values and puts the result on *name*.

    There are 2 ways to use it:

    .. code:: python

        img = ee.Image("LANDSAT/LC8_L1T_TOA_FMASK/LC82310902013344LGN00")
        newimg = Image.sumBands(img, "added_bands", ("B1", "B2", "B3"))

    :param name: name for the band that contains the added values of bands
    :type name: str
    :param bands: names of the bands to be added. If None (default) it sums
        all bands
    :type bands: tuple
    :return: the parsed image with one additional band with the sum of `bands`
    :rtype: ee.Image
    """
    band_names = image.bandNames()
    if bands is None:
        bn = band_names
    else:
        bn = ee.List(list(bands))

    nim = ee.Image(0).select([0], [name])

    # TODO: check if passed band names are in band names # DONE
    def sum_bands(n, ini):
        condition = ee.List(band_names).contains(n)
        return ee.Algorithms.If(condition,
                                ee.Image(ini).add(image.select([n])),
                                ee.Image(ini))

    newimg = ee.Image(bn.iterate(sum_bands, nim))

    return image.addBands(newimg)


def replace(image, to_replace, to_add):
    """ Replace one band of the image with a provided band

    :param to_replace: name of the band to replace. If the image hasn't got
        that band, it will be added to the image.
    :type to_replace: str
    :param to_add: Image (one band) containing the band to add. If an Image
        with more than one band is provided, it uses the first band.
    :type to_add: ee.Image
    :return: Same Image provided with the band replaced
    :rtype: ee.Image
    """
    # TODO: see Image.addBands({overwrite:True})
    band = to_add.select([0])
    bands = image.bandNames()
    resto = bands.remove(to_replace)
    img_resto = image.select(resto)
    img_final = img_resto.addBands(band)
    return img_final


def addConstantBands(image, value=None, *names, **pairs):
    """ Adds bands with a constant value

    :param names: final names for the additional bands
    :type names: str
    :param value: constant value
    :type value: int or float
    :param pairs: keywords for the bands (see example)
    :type pairs: int or float
    :return: the function for ee.ImageCollection.map()
    :rtype: function

    :Example:

    .. code:: python

        from geetools.tools import addConstantBands
        import ee

        col = ee.ImageCollection(ID)

        # Option 1 - arguments
        addC = addConstantBands(0, "a", "b", "c")
        newcol = col.map(addC)

        # Option 2 - keyword arguments
        addC = addConstantBands(a=0, b=1, c=2)
        newcol = col.map(addC)

        # Option 3 - Combining
        addC = addC = addConstantBands(0, "a", "b", "c", d=1, e=2)
        newcol = col.map(addC)
    """
    from functools import reduce

    # check type of value
    # is_val_n = type(value) is int or type(value) is float
    is_val_n = isinstance(value, (int, float))

    if is_val_n and names:
        list1 = [ee.Image.constant(value).select([0], [n]) for n in names]
    else:
        list1 = []

    if pairs:
        list2 = [ee.Image.constant(val).select([0], [key]) for key, val \
                 in pairs.items()]
    else:
        list2 = []

    if list1 or list2:
        lista_img = list1 + list2
    elif value is None:
        raise ValueError("Parameter 'value' must be a number")
    else:
        return addConstantBands(image, value, "constant")

    img_final = reduce(lambda x, y: x.addBands(y), lista_img)

    return ee.Image(image).addBands(ee.Image(img_final))


def minscale(image):
    """ Get the minimal scale of an Image, looking at all Image's bands.
    For example if:
        B1 = 30
        B2 = 60
        B3 = 10
    the function will return 10

    :return: the minimal scale
    :rtype: ee.Number
    """
    bands = image.bandNames()

    first = image.select([ee.String(bands.get(0))])
    ini = ee.Number(first.projection().nominalScale())

    def wrap(name, i):
        i = ee.Number(i)
        scale = ee.Number(image.select([name]).projection().nominalScale())
        condition = scale.lte(i)
        newscale = ee.Algorithms.If(condition, scale, i)
        return newscale

    return ee.Number(bands.slice(1).iterate(wrap, ini))


def mixBands(imgs):
    """ Mix all bands into a single image """
    if isinstance(imgs, (list, tuple)):
        imgs = ee.List(imgs)

    first = ee.Image(imgs.get(0))
    rest = imgs.slice(1)

    return ee.Image(rest.iterate(lambda i, f: ee.Image(f).addBands(i), first))


def computeBits(image, start, end, newName):
    """ Compute the bits of an image

    :param start: start bit
    :type start: int
    :param end: end bit
    :type end: int
    :param newName: new name for the band
    :type newName: str
    :return: A function which single argument is the image and returns a single
        band image of the extracted bits, giving the band a new name
    :rtype: function
    """
    pattern = ee.Number(0)
    start = ee.Number(start).toInt()
    end = ee.Number(end).toInt()
    newName = ee.String(newName)

    seq = ee.List.sequence(start, end)

    def toiterate(element, ini):
        ini = ee.Number(ini)
        bit = ee.Number(2).pow(ee.Number(element))
        return ini.add(bit)

    patt = seq.iterate(toiterate, pattern)

    patt = ee.Number(patt).toInt()

    good_pix = image.select([0], [newName]).toInt() \
        .bitwiseAnd(patt).rightShift(start)
    return good_pix.toInt()


def passProperty(image, to, properties):
    """ Pass properties from one image to another

    :param img_with: image that has the properties to tranpass
    :type img_with: ee.Image
    :param img_without: image that will recieve the properties
    :type img_without: ee.Image
    :param properties: properies to transpass
    :type properties: list
    :return: the image with the new properties
    :rtype: ee.Image
    """
    for prop in properties:
        p = image.get(prop)
        to = to.set(prop, p)
    return to


def goodPix(image, retain=None, drop=None, name='good_pix'):
    """ Get a 'good pixels' bands from the image's bands that retain the good
    pixels and drop the bad pixels. It will first retain the retainable bands
    and then drop the droppable ones

    :param image: the image
    :type image: ee.Image
    :param retain: names of the bands that hold good (want to retain) pixels,
        for example, a good quality band
    :type retain: tuple
    :param drop: names of the bands that hold bad (want to drop) pixels, for
        example a cloud mask band
    :type drop: tuple
    :param name: name for the resulting band
    :type name: str
    :rtype: ee.Image
    """
    to_retain = ee.List(retain)
    to_drop = ee.List(drop)

    def make_or(bandname, ini):
        ini = ee.Image(ini)
        band = image.select(bandname)
        return ini.Or(band)

    final_retain = ee.Image(to_retain.iterate(make_or, empty(0)))
    final_drop = ee.Image(to_drop.iterate(make_or, empty(0)))

    # not bad but not good (retain)
    not_bad_not_good = final_drop.And(final_retain)

    final = not_bad_not_good.bitwiseXor(final_drop)

    return final.select([0], [name])


def toGrid(image, size=1, band=None, geometry=None):
    """ Create a grid from pixels in an image. Results may depend on the image
    projection. Work fine in Landsat imagery.

    IMPORTANT: This grid is not perfect, it can be misplaced and have some
    holes due to projection.

    :param image: the image
    :type image: ee.Image
    :param size: the size of each cell, according to:
        - 1: 1 pixel
        - 2: 9 pixels (3x3)
        - 3: 25 pixels (5x5)
        - and so on..
    :type size: int
    :param band: the band to get the projection (and so, the scale) from. If
        None, the first one will be used
    :type band: str
    :param geometry: the geometry where the grid will be computed. If the image
        is unbounded this parameter must be set in order to work. If None,
        the image geometry will be used if not unbounded.
    :type geometry: ee.Geometry or ee.Feature
    """
    band = band if band else 0
    iband = image.select(band)

    if geometry:
        if isinstance(geometry, ee.Feature):
            geometry = geometry.geometry()
    else:
        geometry = image.geometry()

    projection = iband.projection()
    scale = projection.nominalScale()
    scale = scale.multiply((int(size)*2)-1)
    buffer = scale.divide(2)

    # get coordinates image
    latlon = ee.Image.pixelLonLat().reproject(projection)

    # put each lon lat in a list
    coords = latlon.select(['longitude', 'latitude'])

    coords = coords.reduceRegion(
        reducer= ee.Reducer.toList(),
        geometry= geometry.buffer(scale),
        scale= scale)

    # get lat & lon
    lat = ee.List(coords.get('latitude'))
    lon = ee.List(coords.get('longitude'))

    # zip them. Example: zip([1, 3],[2, 4]) --> [[1, 2], [3,4]]
    point_list = lon.zip(lat)

    def over_list(p):
        p = ee.List(p)
        point = ee.Geometry.Point(p).buffer(buffer).bounds()
        return ee.Feature(point)

    # make grid
    fc = ee.FeatureCollection(point_list.map(over_list))

    return fc


def renamePattern(image, pattern, bands=None):
    """ Rename the bands of the parsed image with the given pattern

    :param image:
    :param pattern: the special keyword `{band}` will be replaced with the
        actual band name. Spaces will be replaced with underscore. It also will
        be trimmed
    :param bands: the bands to rename. If None it'll rename all the bands
    :return:
    """
    allbands = image.bandNames()

    pattern = ee.String(pattern)
    selected = image.select(bands) if bands else image

    pattern = pattern.trim().split(' ').join('_')

    bands_to_replace = selected.bandNames()

    def wrap(name):
        condition = pattern.index('{band}').gt(0)

        return ee.String(ee.Algorithms.If(
            condition,
            pattern.replace('{band}', ee.String(name)),
            ee.String(name)))

    newbands = bands_to_replace.map(wrap)

    new_allbands = ee_list.replaceDict(
        allbands, ee.Dictionary.fromLists(bands_to_replace, newbands))

    return image.select(allbands, new_allbands)


def gaussFunction(image, band, range_min=None, range_max=None, mean=0,
                  std=None, output_min=None, output_max=1, stretch=1,
                  region=None, scale=None, name='gauss', **kwargs):
    """ Apply the Guassian function to an Image.
    https://en.wikipedia.org/wiki/Gaussian_function

    :param band: the name of the band to use
    :type band: str
    :param range_min: the minimum pixel value in the parsed band. If None, it
        will be computed
    :param range_max: the maximum pixel value in the parsed band. If None, it
        will be computed
    :param mean: the position of the center of the peak. Defaults to 0
    :type mean: int or float
    :param std: the standard deviation value. Defaults to range/4
    :type std: int or float
    :param output_max: height of the curve's peak
    :type output_max: int or float
    :param output_min: the desired minimum of the curve
    :type output_min: int or float
    :param stretch: a stretching value. As bigger as stretch
    :type stretch: int or float
    :param name: name of the resulting band
    :type name: strmax
    """
    image = image.select(band)

    if not region:
        region = image.geometry()

    if not scale:
        scale = image.projection().nominalScale()

    if range_min is None and range_max is None:
        minmax = image.reduceRegion(reducer=ee.Reducer.minMax(),
                                    geometry=region, scale=scale, **kwargs)
        minname = '{}_min'.format(band)
        maxname = '{}_max'.format(band)

        range_min = ee.Image.constant(minmax.get(minname))
        range_max = ee.Image.constant(minmax.get(maxname))

    elif range_min is None:
        minmax = image.reduceRegion(reducer=ee.Reducer.min(),
                                    geometry=region, scale=scale, **kwargs)
        range_min = ee.Image.constant(minmax.get(band))
        range_max = castImage(range_max)

    elif range_max is None:
        minmax = image.reduceRegion(reducer=ee.Reducer.max(),
                                    geometry=region, scale=scale, **kwargs)
        range_max = ee.Image.constant(minmax.get(band))
        range_min = castImage(range_min)
    else:
        range_max = castImage(range_max)
        range_min = castImage(range_min)

    mean = castImage(mean)

    if std is None:
        std = range_max.subtract(range_min).divide(4)
    else:
        std = castImage(std)

    output_min = castImage(output_min)
    output_max = castImage(output_max)
    stretch = castImage(stretch)

    def compute_gauss(img):
        result = ee.Image().expression(
            'exp(((val-mean)**2)/(-2*(std**2))*(abs(stretch)))*max',
            {'val': img,
             'mean': mean,
             'std': std,
             'max': output_max,
             'stretch': stretch
             })
        return result

    no_parametrized = compute_gauss(image)

    if output_min is None:
        return no_parametrized.rename(name)
    else:
        min_result = compute_gauss(range_min)
        max_result = compute_gauss(range_max)
        min_result_final = min_result.min(max_result)

        parametrized = ee.Image().expression(
            '(value-min_result)/(max-min_result)*(max-min)+min',
            {'value': no_parametrized,
             'min_result': min_result_final,
             'max': output_max,
             'min': output_min
             })
        return parametrized.rename(name)


def makeName(img, pattern, date_pattern=None, extra=None):
    """ Make a name with the given pattern. The pattern must contain the
    propeties to replace between curly braces. There are 2 special words:

    * 'system_date': replace with the date of the image formatted with
      `date_pattern`, which defaults to 'yyyyMMdd'
    * 'id' or 'ID': the image id. If None, it'll be replaced with 'id'

    Pattern example (supposing each image has a property called `city`):
    'image from {city} on {system_date}'

    You can add extra parameters using keyword `extra`
    """
    img = ee.Image(img)
    props = img.toDictionary()
    props = ee.Dictionary(ee.Algorithms.If(
        img.id(),
        props.set('id', img.id()).set('ID', img.id()),
        props))
    props = ee.Dictionary(ee.Algorithms.If(
        img.propertyNames().contains('system:time_start'),
        props.set('system_date', img.date().format(date_pattern)),
        props))
    if extra:
        extra = ee.Dictionary(extra)
        props = props.combine(extra)
    name = string.format(pattern, props)

    return name


def normalDistribution(image, band, mean=None, std=None, region=None,
                       scale=None, name='normal_distribution', **kwargs):
    """ Compute a Normal Distribution using the Gaussian Function """
    pi = ee.Number(math.pi)

    image = image.select(band)

    if mean is None:
        mean = image.reduceRegion(reducer=ee.Reducer.mean(),
                                  geometry=region, scale=scale, **kwargs)
        mean = ee.Image.constant(mean.get(band))
    else:
        mean = castImage(mean)

    if std is None:
        std = image.reduceRegion(reducer=ee.Reducer.stdDev(),
                                 geometry=region, scale=scale, **kwargs)
        std = ee.Image.constant(std.get(band))
    else:
        std = castImage(std)

    output_max = ee.Image(1)\
                   .divide(std.multiply(ee.Image(2).multiply(pi).sqrt()))

    return gaussFunction(image, band, mean=mean, std=std,
                         output_max=output_max, name=name, **kwargs)


def linearFunction(image, band, range_min=None, range_max=None, mean=None,
                   output_min=None, output_max=None, name='linear_function',
                   region=None, scale=None, **kwargs):
    """ Apply a linear function over one image band using the following
    formula:

    - a = abs(val-mean)
    - b = output_max-output_min
    - c = abs(range_max-mean)
    - d = abs(range_min-mean)
    - e = max(c, d)

    f(x) = a*(-1)*(b/e)+output_max

    :param band: the band to process
    :param range_min: the minimum pixel value in the parsed band. If None, it
        will be computed over the parsed region (heavy process that can fail)
    :param range_max: the maximum pixel value in the parsed band. If None, it
        will be computed over the parsed region (heavy process that can fail)
    :param output_min: the minimum value that will take the resulting band.
    :param output_max: the minimum value that will take the resulting band.
    :param mean: the value on the given range that will take the `output_max`
        value
    :param name: the name of the resulting band
    :param region: the region to reduce over if no `range_min` and/or no
        `range_max` has been parsed
    :param scale: the scale that will be use for reduction if no `range_min`
        and/or no `range_max` has been parsed
    :param kwargs: extra arguments for the reduction: crs, crsTransform,
        bestEffort, maxPixels, tileScale.
    :return: a one band image that results of applying the linear function
        over every pixel in the image
    :rtype: ee.Image
    """
    image = image.select(band)

    if not region:
        region = image.geometry()

    if not scale:
        scale = image.projection().nominalScale()

    if range_min is None and range_max is None:
        minmax = image.reduceRegion(reducer=ee.Reducer.minMax(),
                                    geometry=region, scale=scale, **kwargs)
        minname = '{}_min'.format(band)
        maxname = '{}_max'.format(band)

        imin = ee.Image.constant(minmax.get(minname))
        imax = ee.Image.constant(minmax.get(maxname))

    elif range_min is None:
        minmax = image.reduceRegion(reducer=ee.Reducer.min(),
                                    geometry=region, scale=scale, **kwargs)
        imin = ee.Image.constant(minmax.get(band))
        imax = castImage(range_max)

    elif range_max is None:
        minmax = image.reduceRegion(reducer=ee.Reducer.max(),
                                    geometry=region, scale=scale, **kwargs)
        imax = ee.Image.constant(minmax.get(band))
        imin = castImage(range_min)
    else:
        imax = castImage(range_max)
        imin = castImage(range_min)

    if mean is None:
        imean = imax
    else:
        imean = castImage(mean)

    if output_max is None:
        output_max = imax

    if output_min is None:
        output_min = imin

    a = imax.subtract(imean).abs()
    b = imin.subtract(imean).abs()
    t = a.max(b)

    result = ee.Image().expression(
        'abs(val-mean)*(-1)*((max-min)/t)+max',
        {'val': image,
         'mean': imean,
         't': t,
         'imin': imin,
         'max': output_max,
         'min': output_min
         })

    return result.rename(name)


def doyToDate(image, dateFormat='yyyyMMdd', year=None):
    """ Make a date band from a day of year band """
    if not year:
        year = image.date().get('year')

    doyband = image.select([0])
    leap = date.isLeap(year)
    limit = ee.Number(ee.Algorithms.If(leap, 365, 364))
    alldoys = ee.List.sequence(1, limit)

    def wrap(doy, i):
        i = ee.Image(i)
        doy = ee.Number(doy)
        d = date.fromDOY(doy, year)
        date_band = ee.Image.constant(ee.Number.parse(d.format(dateFormat)))
        condition = i.eq(doy)
        return i.where(condition, date_band)

    datei = ee.Image(alldoys.iterate(wrap, doyband))

    return datei.rename('date')


def maskInside(image, geometry):
    """ This is the opposite to ee.Image.clip(geometry) """
    mask = ee.Image.constant(1).clip(geometry).mask().Not()
    return image.updateMask(mask)


def paint(image, featurecollection, vis_params=None, color='black', width=1,
          fillColor=None, **kwargs):
    """ Paint a FeatureCollection onto an Image. Returns an Image with three
    bands: vis-blue, vis-geen, vis-red (uint8)

    It admits the same parameters as ee.FeatureCollection.style
    """
    if not fillColor:
        fillColor = "#00000000"
    if not vis_params:
        firstband = ee.String(image.bandNames().get(0))
        vis_params = dict(bands=[firstband, firstband, firstband], min=0,
                          max=1)
    region = image.geometry()
    filtered = ee.FeatureCollection(
        ee.Algorithms.If(
            region.isUnbounded(),
            featurecollection,
            featurecollection.filterBounds(region)
        ))
    fcraster = filtered.style(color=color, width=width, fillColor=fillColor,
                              **kwargs)
    mask = fcraster.reduce('sum').gte(0).rename('mask')
    topaint = image.visualize(**vis_params)
    final = topaint.where(mask, fcraster)
    final = final.copyProperties(source=image)
    properties = image.propertyNames()
    final = ee.Image(ee.Algorithms.If(
        properties.contains('system:time_start'),
        final.set('system:time_start', image.date().millis()),
        final))
    final = ee.Image(ee.Algorithms.If(
        properties.contains('system:time_end'),
        final.set('system:time_end', ee.Number(image.get('system:time_end'))),
        final))

    return final


def repeatBand(image, times=None, names=None, properties=None):
    """ Repeat one band. If the image parsed has more than one band, the first
    will be used """
    band = ee.Image(image.select([0]))
    if times is not None:
        times = ee.Number(times)
        proxylist = ee.List.repeat(0, times.subtract(1))
        def add(band, i):
            band = ee.Image(band)
            i = ee.Image(i)
            return i.addBands(band)
        proxyImg = proxylist.map(lambda n: band)
        repeated = ee.Image(proxyImg.iterate(add, band))
    else:
        newNames = ee.List(names)
        firstName = ee.String(newNames.get(0))
        rest = ee.List(newNames.slice(1))
        def add(name, i):
            name = ee.String(name)
            i = ee.Image(i)
            return i.addBands(band.rename(name))
        first = band.rename(firstName)
        repeated = ee.Image(rest.iterate(add, first))

    if properties:
        repeated = repeated.setMulti(properties)

    return ee.Image(repeated)


def arrayNonZeros(image):
    """
    Return an image array without zeros

    :param image:
    :return:
    """
    def wrap(arr):
        binarr = arr.divide(arr)
        n = binarr.arrayReduce(ee.Reducer.sum(), [0]).multiply(-1)
        nimg = n.arrayProject([0]).arrayFlatten([['n']]).toInt()
        sorted = arr.arraySort()
        sliced = sorted.arraySlice(0, nimg)
        return sliced

    bands = image.bandNames()
    first = wrap(image.select([bands.get(0)]))
    rest = bands.slice(1)

    def overBands(band, i):
        band = ee.String(band)
        i = ee.Image(i)
        array = wrap(image.select([band]))
        return i.addBands(array)

    result1 = ee.Image(rest.iterate(overBands, first))
    result2 = ee.Image(wrap(image))

    return ee.Image(ee.Algorithms.If(bands.size(), result1, result2))


def getTileURL(image, visParams=None):
    """ Get the URL for the given image passing a normal visualization
    parameters like `{'bands':['B4','B3','B2'], 'min':0, 'max':5000}` """
    if visParams:
        vis = mapui.formatVisParams(visParams)
        image_info = image.getMapId(vis)
    else:
        image_info = image.getMapId()
    fetcher = image_info['tile_fetcher']
    return fetcher.url_format


def applyMask(image, mask, bands=None, negative=True):
    """ Apply a passed positive mask """
    bands = bands or mask.bandNames()
    bands = ee.List(bands)
    def wrap(band, img):
        img = ee.Image(img)
        band = ee.String(band)
        m = mask.select(band)
        toapply = m.Not() if negative else m
        return img.updateMask(toapply)

    return ee.Image(bands.iterate(wrap, image))


def maskCover(image, geometry=None, scale=None, property_name='MASK_COVER',
              crs=None, crsTransform=None, bestEffort=False,
              maxPixels=1e13, tileScale=1):
    """ Percentage of masked pixels (masked/total * 100) as an Image property

    :param image: ee.Image holding the mask. If the image has more than
        one band, the first one will be used
    :type image: ee.Image
    :param geometry: the value will be computed inside this geometry. If None,
        will use image boundaries. If unbounded the result will be 0
    :type geometry: ee.Geometry or ee.Feature
    :param scale: the scale of the mask
    :type scale: int
    :param property_name: the name of the resulting property
    :type property_name: str
    :return: The same parsed image with a new property holding the mask cover
        percentage
    :rtype: ee.Image
    """
    # keep only first band
    imageband = image.select(0)

    # get projection
    projection = imageband.projection()

    if not scale:
        scale = projection.nominalScale()

    # get band name
    band = ee.String(imageband.bandNames().get(0))

    # Make an image with all ones
    ones_i = ee.Image.constant(1).reproject(projection).rename(band)

    if not geometry:
        geometry = image.geometry()

    # manage geometry types
    if isinstance(geometry, (ee.Feature, ee.FeatureCollection)):
        geometry = geometry.geometry()

    unbounded = geometry.isUnbounded()

    # Get total number of pixels
    ones = ones_i.reduceRegion(
        reducer= ee.Reducer.count(),
        geometry= geometry,
        scale= scale,
        maxPixels= maxPixels,
        crs=crs,
        crsTransform=crsTransform,
        bestEffort=bestEffort,
        tileScale=tileScale,
    ).get(band)
    ones = ee.Number(ones)

    # select first band, unmask and get the inverse
    mask = imageband.mask()
    mask_not = mask.Not()
    image_to_compute = mask.updateMask(mask_not)

    # Get number of zeros in the given image
    zeros_in_mask =  image_to_compute.reduceRegion(
        reducer= ee.Reducer.count(),
        geometry= geometry,
        scale= scale,
        maxPixels= maxPixels,
        crs=crs,
        crsTransform=crsTransform,
        bestEffort=bestEffort,
        tileScale=tileScale
    ).get(band)
    zeros_in_mask = ee.Number(zeros_in_mask)

    percentage = zeros_in_mask.divide(ones)
    # Multiply by 100
    cover = percentage.multiply(100)

    # Return None if geometry is unbounded
    final = ee.Number(ee.Algorithms.If(unbounded, 0, cover))

    return image.set(property_name, final)


def regionCover(image, region, bands=None, scale=None, operator='OR',
                property_name='REGION_COVER', crs=None, crsTransform=None,
                bestEffort=False, maxPixels=1e13, tileScale=1):
    """ Compute the percentage of values greater than 1 in a region. If more
     than one band is specified, it applies the specified operator """
    operators = ['OR', 'AND']
    if operator not in operators:
        raise ValueError('operator must be one of {}'.format(operators))

    if not bands:
        bands = ee.List([image.bandNames().get(0)])

    if operator == 'AND':
        reducer = ee.Reducer.bitwiseAnd()
    else:
        reducer = ee.Reducer.bitwiseOr()

    bandname = 'regionCover'
    mask = image.select(bands).reduce(reducer).rename(bandname)

    # get projection
    projection = mask.projection()

    if not scale:
        scale = projection.nominalScale()

    # Make an image with all ones
    ones_i = ee.Image.constant(1).reproject(projection).rename(bandname)

    # manage geometry types
    if isinstance(region, (ee.Feature, ee.FeatureCollection)):
        region = region.geometry()

    unbounded = region.isUnbounded()

    # Get total number of pixels
    ones = ones_i.reduceRegion(
        reducer= ee.Reducer.count(),
        geometry= region,
        scale= scale,
        maxPixels= maxPixels,
        crs=crs,
        crsTransform=crsTransform,
        bestEffort=bestEffort,
        tileScale=tileScale,
    ).get(bandname)
    ones = ee.Number(ones)

    # select first band, unmask and get the inverse
    image_to_compute = mask.selfMask()

    # Get number of zeros in the given image
    zeros_in_mask =  image_to_compute.reduceRegion(
        reducer= ee.Reducer.count(),
        geometry= region,
        scale= scale,
        maxPixels= maxPixels,
        crs=crs,
        crsTransform=crsTransform,
        bestEffort=bestEffort,
        tileScale=tileScale
    ).get(bandname)
    zeros_in_mask = ee.Number(zeros_in_mask)

    percentage = zeros_in_mask.divide(ones)

    # Multiply by 100
    cover = percentage.multiply(100)

    # Return None if geometry is unbounded
    final = ee.Number(ee.Algorithms.If(unbounded, 0, cover))
    return image.set(property_name, final)


def proxy(values=(0,), names=('constant',), types=('int8',)):
    """ Create a proxy image with the given values, names and types

    :param values: list of values for every band of the resulting image
    :type values: list
    :param names: list of names
    :type names: list
    :param types: list of band types. Options are  'int8', 'int16', 'int32',
    'int64', 'uint8', 'uint16', 'uint32', 'byte', 'short', 'int', 'long',
    'float' and 'double'
    :type types: list
    :rtype: ee.Image
    """
    values = list(values)
    names = list(names)
    types = list(types)
    tps = dict(zip(names, types))
    im = ee.Image(values).rename(names).cast(tps)
    # update mask
    for v, n in zip(values, names):
        if v is None:
            band = im.select(n)
            masked = band.selfMask()
            im = im.addBands(masked, overwrite=True)
    return im


def clipToCollection(image, featureCollection, keepFeatureProperties=True):
    """ Clip an image using each feature of a collection and return an
    ImageCollection with one image per feature """
    def overFC(feat):
        geom = feat.geometry()
        clipped = image.clip(geom)
        if keepFeatureProperties:
            clipped = clipped.copyProperties(feat)
        return clipped

    ic = ee.ImageCollection(featureCollection.map(overFC))
    return ic


class Classification(object):
    """ Class holding (static) methods for classified images. """
    @staticmethod
    def vectorize(image, categories, label='label'):
        """ Reduce to vectors the selected classes fro a classified image

        :param categories: the categories to vectorize
        :type categories: list

        """
        def over_cat(cat, ini):
            cat = ee.Number(cat)
            ini = ee.Image(ini)
            return ini.add(image.eq(cat).multiply(cat))

        filtered = ee.Image(
            ee.List(categories).iterate(over_cat,
                                        empty(0, [label])))

        out = filtered.neq(0)
        filtered = filtered.updateMask(out)

        return filtered.reduceToVectors(**{
            'scale': 30,
            'maxPixels':1e13,
            'labelProperty': label})



""" Module for classification tools """
import ee


def binaryRasterAccuracy(truth, classified, region=None):
    """ Get a class raster with the following classes:

        band "classes":
        0: no change detected and no real change (true negative)
        1: no change detected but real change (false negative)
        2: change detected but not real change (false positive)
        3: change detected and real change (true positive)

        band "truth":
        0: no change
        1: change

        source: https://en.wikipedia.org/wiki/Evaluation_of_binary_classifiers
        """
    # convert to int
    truth = truth.select([0]).toInt().unmask()
    classified = classified.select([0]).toInt().unmask()

    if region:
        truth = truth.clip(region)
        classified = classified.clip(region)

    final = ee.Image().expression("""
        t == 1 && c == 1 ? 3 :
        t == 0 && c == 1 ? 2 :
        t == 1 && c == 0 ? 1 : 0
        """, dict(t=truth, c=classified)).rename('classes')

    return final


def binaryMetrics(truth, classified, scale, region=None):
    """ get accuracy from a truth image and a classified image

    names from: https://en.wikipedia.org/wiki/Evaluation_of_binary_classifiers

    :param truth: the truth raster. Only the first band will be used
    :type truth: ee.Image
    :param classified: the classification results. Only the first band will be
        used
    :type classified: ee.Image
    :param scale: the scale for the analysis
    :type scale: int
    :param region: the region for the analysis.
    :type region: ee.Geometry
    :return: a confusion matrix
    :rtype: ee.ConfusionMatrix
    """
    class_list = ee.List(['0', '1', '2', '3'])
    names_list = ee.List(['TN', 'FN', 'FP', 'TP'])

    scale = int(scale)

    classes = binaryRasterAccuracy(truth, classified, region)

    params = dict(
        reducer=ee.Reducer.frequencyHistogram(),
        scale=scale,
        maxPixels=int(1e13),
        tileScale=4
    )
    if region:
        params['geometry'] = region

    # get frecuency histogram (classes)
    matrix = classes.reduceRegion(**params)
    matrix = ee.Dictionary(matrix.get('classes'))  # cast

    def fillDict(clas, d):
        d = ee.Dictionary(d)
        clas = ee.String(clas)
        cond = d.contains(clas)
        return ee.Dictionary(ee.Algorithms.If(cond, d, d.set(clas, 0)))

    class_dict = ee.Dictionary(
        class_list.iterate(fillDict, matrix)).rename(class_list, names_list)

    TP = ee.Number(class_dict.get('TP'))
    TN = ee.Number(class_dict.get('TN'))
    FP = ee.Number(class_dict.get('FP'))
    FN = ee.Number(class_dict.get('FN'))

    ALL = dict(TP=TP, TN=TN, FP=FP, FN=FN)

    # Metrics
    TPR = ee.Number.expression("TP/(TP+FN)", ALL)
    TNR = ee.Number.expression("TN/(TN+FP)", ALL)
    PPV = ee.Number.expression("TP/(TP+FP)", ALL)
    NPV = ee.Number.expression("TN/(TN+FN)", ALL)
    FNR = ee.Number.expression("FN/(FN+TP)", ALL)
    FPR = ee.Number.expression("FP/(FP+TN)", ALL)
    FDR = ee.Number.expression("FP/(FP+TP)", ALL)
    FOR = ee.Number.expression("FN/(FN+TN)", ALL)
    TS = ee.Number.expression("TP/(TP+FN+FP)", ALL)
    ACC = ee.Number.expression("(TP+TN)/(TP+TN+FP+FN)", ALL)
    BA = ee.Number.expression("(TPR+TNR)/2", dict(TPR=TPR, TNR=TNR))
    F1 = ee.Number.expression("(2*TP)/(2*TP+FP+FN)", ALL)

    metrics = ee.Dictionary(dict(
        TPR=TPR, TNR=TNR, PPV=PPV, NPV=NPV, FNR=FNR,
        FPR=FPR, FDR=FDR, FOR=FOR, TS=TS, ACC=ACC,
        BA=BA, F1=F1
    ))

    return class_dict.combine(metrics)
